import functools
import random
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

from autoblocks._impl.prompts.placeholders import TemplatePlaceholder
from autoblocks._impl.prompts.placeholders import parse_placeholders_from_template

try:
    import pydantic

    assert pydantic.__version__.startswith("2.")
except (ImportError, AssertionError):
    raise ImportError(
        "The Autoblocks prompt SDK requires pydantic version 2.x. "
        "You can install it with `pip install pydantic==2.*`."
    )


class FrozenModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(frozen=True)


class PromptParams(FrozenModel):
    """LLM model parameters for a prompt."""

    params: Optional[Dict[str, Any]]


class PromptTemplate(FrozenModel):
    """An individual template for a prompt."""

    id: str
    template: str

    @functools.cached_property
    def placeholders(self) -> List[TemplatePlaceholder]:
        return parse_placeholders_from_template(self.template)


class Prompt(FrozenModel):
    """A schema representing a prompt managed in the Autoblocks platform."""

    id: str
    version: str
    revision_id: str = pydantic.Field(..., alias="revisionId")
    params: Optional[PromptParams] = None
    templates: List[PromptTemplate]

    @functools.cached_property
    def tracking(self) -> dict[str, Any]:
        """
        Returns the schema expected by our ingestion endpoint's prompt tracking property.
        """
        return self.model_dump(by_alias=True)


class WeightedMinorVersion(FrozenModel):
    version: str
    weight: float = pydantic.Field(..., gt=0)


class PromptMinorVersion(FrozenModel):
    version: Union[str, List[WeightedMinorVersion]]

    @property
    def str_version(self) -> Optional[str]:
        if isinstance(self.version, list):
            return None
        return self.version

    @property
    def weighted_versions(self) -> Optional[List[WeightedMinorVersion]]:
        if isinstance(self.version, list):
            return self.version
        return None

    @functools.cached_property
    def all_minor_versions(self) -> Set[str]:
        versions: Set[str] = set()
        if weighted_versions := self.weighted_versions:
            for v in weighted_versions:
                versions.add(v.version)
        elif str_version := self.str_version:
            versions.add(str_version)
        return versions

    def choose_version(self) -> str:
        if weighted_versions := self.weighted_versions:
            # Minor version is a weighted list; choose one according
            # to their weights.
            (rand_str_version,) = random.choices(
                population=[v.version for v in weighted_versions],
                weights=[v.weight for v in weighted_versions],
                k=1,
            )
            return rand_str_version
        elif str_version := self.str_version:
            # Minor version is a constant; return it.
            return str_version

        raise RuntimeError("Minor version should either be a string or a list of weighted versions.")


class AutogeneratePromptConfig(FrozenModel):
    model_config = pydantic.ConfigDict(coerce_numbers_to_str=True)

    id: str
    major_version: Optional[str] = None
    dangerously_use_undeployed_revision: Optional[str] = None


class AutogeneratePromptsConfig(FrozenModel):
    outfile: str
    prompts: List[AutogeneratePromptConfig]



import logging
from dataclasses import asdict
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import httpx

from autoblocks._impl.api.models import AbsoluteTimeFilter
from autoblocks._impl.api.models import Event
from autoblocks._impl.api.models import ManagedTestCase
from autoblocks._impl.api.models import ManagedTestCaseResponse
from autoblocks._impl.api.models import RelativeTimeFilter
from autoblocks._impl.api.models import Trace
from autoblocks._impl.api.models import TraceFilter
from autoblocks._impl.api.models import TracesResponse
from autoblocks._impl.api.models import View
from autoblocks._impl.config.constants import API_ENDPOINT
from autoblocks._impl.util import AutoblocksEnvVar

log = logging.getLogger(__name__)


def make_trace_response(data: Dict[str, Any]) -> Trace:
    return Trace(
        id=data["id"],
        events=[
            Event(
                id=event["id"],
                trace_id=event["traceId"],
                message=event["message"],
                timestamp=event["timestamp"],
                properties=event.get("properties") or {},
            )
            for event in data["events"]
        ],
    )


def make_traces_response(data: Dict[str, Any]) -> TracesResponse:
    return TracesResponse(
        next_cursor=data.get("nextCursor"),
        traces=[make_trace_response(trace) for trace in data["traces"]],
    )


def snake_to_camel(s: str) -> str:
    return "".join(word.lower() if i == 0 else word.capitalize() for i, word in enumerate(s.split("_")))


def camel_case_factory(values: List[Tuple[str, Any]]) -> Dict[str, Any]:
    return dict(((snake_to_camel(k)), v) for k, v in values if v is not None)


class AutoblocksAPIClient:
    def __init__(self, api_key: Optional[str] = None, timeout: timedelta = timedelta(seconds=10)) -> None:
        api_key = api_key or AutoblocksEnvVar.API_KEY.get()
        if not api_key:
            raise ValueError(f"You must provide an api_key or set the {AutoblocksEnvVar.API_KEY} environment variable.")
        self._client = httpx.Client(
            base_url=API_ENDPOINT,
            headers={"Authorization": f"Bearer {api_key}"},
            timeout=timeout.total_seconds(),
        )

    def get_views(self) -> List[View]:
        req = self._client.get("/views")
        req.raise_for_status()
        resp = req.json()
        return [View(id=view["id"], name=view["name"]) for view in resp]

    def get_traces_from_view(self, view_id: str, *, page_size: int, cursor: Optional[str] = None) -> TracesResponse:
        req = self._client.get(
            f"/views/{view_id}/traces",
            params={"pageSize": page_size, "cursor": cursor or ""},
        )
        req.raise_for_status()
        return make_traces_response(req.json())

    def search_traces(
        self,
        *,
        page_size: int,
        time_filter: Union[RelativeTimeFilter, AbsoluteTimeFilter],
        trace_filters: Optional[List[TraceFilter]] = None,
        query: Optional[str] = None,
        cursor: Optional[str] = None,
    ) -> TracesResponse:
        payload = dict(
            pageSize=page_size,
            timeFilter=asdict(time_filter, dict_factory=camel_case_factory) if time_filter else None,
            traceFilters=[asdict(f, dict_factory=camel_case_factory) for f in trace_filters] if trace_filters else None,
            query=query,
            cursor=cursor,
        )
        payload = {k: v for k, v in payload.items() if v is not None}
        req = self._client.post(
            "/traces/search",
            json=payload,
        )
        req.raise_for_status()
        return make_traces_response(req.json())

    def get_test_cases(self, test_suite_id: str) -> ManagedTestCaseResponse:
        req = self._client.get(f"/test-suites/{test_suite_id}/test-cases")
        req.raise_for_status()
        resp = req.json()
        return ManagedTestCaseResponse(
            test_cases=[ManagedTestCase(id=case["id"], body=case["body"]) for case in resp["testCases"]]
        )



import dataclasses
import hashlib
import itertools
import traceback
from typing import Any
from typing import Generator
from typing import Optional
from typing import Sequence

import orjson

from autoblocks._impl.testing.models import BaseTestCase
from autoblocks._impl.testing.models import HumanReviewField
from autoblocks._impl.testing.models import TestCaseConfig
from autoblocks._impl.testing.models import TestCaseContext
from autoblocks._impl.testing.models import TestCaseType

# This attribute name might sound redundant, but it is named this
# way (as opposed to just `config`) to decrease the likelihood
# our config attr name conflicts with the name of an attr the user
# wants to use on their test case.
TEST_CASE_CONFIG_ATTR = "test_case_config"


def md5(text: str) -> str:
    return hashlib.md5(text.encode()).hexdigest()


def orjson_default(o: Any) -> Any:
    if hasattr(o, "model_dump_json") and callable(o.model_dump_json):
        # pydantic v2
        return orjson.loads(o.model_dump_json())
    elif hasattr(o, "json") and callable(o.json):
        # pydantic v1
        return orjson.loads(o.json())
    elif isinstance(o, Exception):
        return "".join(
            traceback.format_exception(
                type(o),
                o,
                o.__traceback__,
            )
        )
    raise TypeError


def serialize(x: Any) -> Any:
    return orjson.loads(orjson.dumps(x, default=orjson_default))


def serialize_test_case(test_case: BaseTestCase) -> Any:
    # See https://docs.python.org/3/library/dataclasses.html#dataclasses.is_dataclass:
    # isinstance(test_case, type) checks test_case is an instance and not a type
    if dataclasses.is_dataclass(test_case) and not isinstance(test_case, type):
        serialized: dict[Any, Any] = {}
        for k, v in dataclasses.asdict(test_case).items():
            if k == TEST_CASE_CONFIG_ATTR:
                # Don't serialize the config
                continue
            try:
                serialized[k] = serialize(v)
            except Exception:
                # Skip over non-serializable test case attributes
                pass
        return serialized

    return serialize(test_case)


def serialize_human_review_fields(fields: Optional[list[HumanReviewField]]) -> Optional[list[dict[str, str]]]:
    if fields is not None:
        return [f.serialize() for f in fields]
    return None


def serialize_test_case_for_human_review(test_case: BaseTestCase) -> Optional[list[dict[str, str]]]:
    return serialize_human_review_fields(test_case.serialize_for_human_review())


def serialize_output_for_human_review(output: Any) -> Optional[list[dict[str, str]]]:
    if callable(getattr(output, "serialize_for_human_review", None)):
        return serialize_human_review_fields(output.serialize_for_human_review())
    return None


def config_from_test_case(test_case: TestCaseType) -> Optional[TestCaseConfig]:
    config = getattr(test_case, TEST_CASE_CONFIG_ATTR, None)
    if isinstance(config, TestCaseConfig):
        return config
    return None


def yield_test_case_contexts_from_test_cases(
    test_cases: Sequence[TestCaseType],
) -> Generator[TestCaseContext[TestCaseType], None, None]:
    for test_case in test_cases:
        config = config_from_test_case(test_case)
        if not config or config.repeat_num_times is None:
            yield TestCaseContext(
                test_case=test_case,
                repetition_idx=None,
            )
        else:
            for idx in range(config.repeat_num_times):
                yield TestCaseContext(
                    test_case=test_case,
                    repetition_idx=idx,
                )


GridSearchParams = dict[str, Sequence[Any]]
GridSearchParamsCombo = dict[str, Any]


def yield_grid_search_param_combos(params: GridSearchParams) -> Generator[GridSearchParamsCombo, None, None]:
    keys = list(params.keys())
    values = list(params.values())
    for combo in itertools.product(*values):
        yield dict(zip(keys, combo))



import asyncio
import contextvars
import dataclasses
import functools
import inspect
import json
import logging
import os
import time
import traceback
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import overload

from httpx import HTTPStatusError
from httpx import Response

from autoblocks._impl import global_state
from autoblocks._impl.context_vars import EvaluatorRunContext
from autoblocks._impl.context_vars import TestCaseRunContext
from autoblocks._impl.context_vars import evaluator_run_context_var
from autoblocks._impl.context_vars import get_revision_usage
from autoblocks._impl.context_vars import grid_search_context_var
from autoblocks._impl.context_vars import test_case_run_context_var
from autoblocks._impl.testing.models import BaseTestCase
from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import TestCaseContext
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.util import GridSearchParams
from autoblocks._impl.testing.util import GridSearchParamsCombo
from autoblocks._impl.testing.util import serialize
from autoblocks._impl.testing.util import serialize_output_for_human_review
from autoblocks._impl.testing.util import serialize_test_case
from autoblocks._impl.testing.util import serialize_test_case_for_human_review
from autoblocks._impl.testing.util import yield_grid_search_param_combos
from autoblocks._impl.testing.util import yield_test_case_contexts_from_test_cases
from autoblocks._impl.util import AutoblocksEnvVar
from autoblocks._impl.util import all_settled
from autoblocks.tracer import flush

log = logging.getLogger(__name__)

test_case_semaphore_registry: dict[str, asyncio.Semaphore] = {}  # test_id -> semaphore
evaluator_semaphore_registry: dict[str, dict[str, asyncio.Semaphore]] = {}  # test_id -> evaluator_id -> semaphore

DEFAULT_MAX_TEST_CASE_CONCURRENCY = 10


def tests_and_hashes_overrides_map() -> Optional[dict[str, list[str]]]:
    """
    AUTOBLOCKS_OVERRIDES_TESTS_AND_HASHES is a JSON string that maps test suite IDs to a list of test case hashes.
    This is set when a user triggers a test run from the UI so that we only run the given test suite, and,
    if applicable, only the given test cases.
    """
    raw = AutoblocksEnvVar.OVERRIDES_TESTS_AND_HASHES.get()
    if not raw:
        return None
    return json.loads(raw)  # type: ignore


async def post_to_cli(
    path: str,
    json: dict[str, Any],
) -> Optional[Response]:
    cli_server_address = AutoblocksEnvVar.CLI_SERVER_ADDRESS.get()
    if cli_server_address:
        return await global_state.http_client().post(
            f"{cli_server_address}{path}",
            json=json,
            timeout=30,  # seconds
        )

    # If the CLI server address is not set then the tests are being run
    # directly. In this case we just log the request that would have been
    # sent to the CLI.
    log.debug(f"{path}: {json}")
    return None


async def send_error(
    test_id: str,
    run_id: Optional[str],
    test_case_hash: Optional[str],
    evaluator_id: Optional[str],
    error: Exception,
) -> None:
    await post_to_cli(
        "/errors",
        json=dict(
            testExternalId=test_id,
            runId=run_id,
            testCaseHash=test_case_hash,
            evaluatorExternalId=evaluator_id,
            error=dict(
                name=type(error).__name__,
                message=str(error),
                stacktrace=traceback.format_exc(),
            ),
        ),
    )


async def run_evaluator_unsafe(
    test_id: str,
    run_id: Optional[str],
    test_case_ctx: TestCaseContext[TestCaseType],
    output: Any,
    hook_results: Any,
    evaluator: BaseTestEvaluator,
) -> None:
    """
    This is suffixed with _unsafe because it doesn't handle exceptions.
    Its caller will catch and handle all exceptions.
    """
    async with evaluator_semaphore_registry[test_id][evaluator.id]:
        if hook_results is not None:
            kwargs = dict(hook_results=hook_results)
        else:
            kwargs = dict()

        if inspect.iscoroutinefunction(evaluator.evaluate_test_case):
            evaluation = await evaluator.evaluate_test_case(test_case_ctx.test_case, output, **kwargs)
        else:
            ctx = contextvars.copy_context()
            evaluation = await global_state.event_loop().run_in_executor(
                None,
                ctx.run,
                functools.partial(
                    evaluator.evaluate_test_case,
                    test_case_ctx.test_case,
                    output,
                    **kwargs,
                ),
            )

    if evaluation is None:
        return

    # Revision usage is collected throughout an evaluator's evaluate_test_case call on a test case
    revision_usage = get_revision_usage()

    await post_to_cli(
        "/evals",
        json=dict(
            testExternalId=test_id,
            runId=run_id,
            testCaseHash=test_case_ctx.hash(),
            evaluatorExternalId=evaluator.id,
            score=evaluation.score,
            threshold=dataclasses.asdict(evaluation.threshold) if evaluation.threshold else None,
            metadata=evaluation.metadata,
            revisionUsage=[usage.serialize() for usage in revision_usage] if revision_usage else None,
        ),
    )


async def run_evaluator(
    test_id: str,
    run_id: Optional[str],
    test_case_ctx: TestCaseContext[TestCaseType],
    output: Any,
    hook_results: Any,
    evaluator: BaseTestEvaluator,
) -> None:
    reset_token = evaluator_run_context_var.set(
        EvaluatorRunContext(),
    )
    try:
        await run_evaluator_unsafe(
            test_id=test_id,
            run_id=run_id,
            test_case_ctx=test_case_ctx,
            output=output,
            hook_results=hook_results,
            evaluator=evaluator,
        )
    except Exception as err:
        await send_error(
            test_id=test_id,
            run_id=run_id,
            test_case_hash=test_case_ctx.hash(),
            evaluator_id=evaluator.id,
            error=err,
        )
    finally:
        evaluator_run_context_var.reset(reset_token)


async def run_test_case_unsafe(
    test_id: str,
    run_id: Optional[str],
    test_case_ctx: TestCaseContext[TestCaseType],
    fn: Union[Callable[[TestCaseType], Any], Callable[[TestCaseType], Awaitable[Any]]],
    before_evaluators_hook: Optional[Callable[[TestCaseType, Any], Any]],
) -> Tuple[Any, Any]:
    """
    This is suffixed with _unsafe because it doesn't handle exceptions.
    Its caller will catch and handle all exceptions.
    """
    async with test_case_semaphore_registry[test_id]:
        # NOTE: This should be _inside_ the `async with` block to ensure we don't start the
        # timer until the semaphore is acquired.
        start_time = time.perf_counter()

        if inspect.iscoroutinefunction(fn):
            output = await fn(test_case_ctx.test_case)
        else:
            ctx = contextvars.copy_context()
            output = await global_state.event_loop().run_in_executor(
                None,
                ctx.run,
                fn,
                test_case_ctx.test_case,
            )

        # Calculate duration before running hooks so that the duration only
        # includes time spent in fn()
        test_case_duration_ms = (time.perf_counter() - start_time) * 1_000

        # Run the before-evaluators hook if provided.
        # Note we run this within the test case semaphore so that
        # max_test_case_concurrency applies to both fn + before_evaluators_hook.
        hook_results = None
        if before_evaluators_hook:
            if inspect.iscoroutinefunction(before_evaluators_hook):
                hook_results = await before_evaluators_hook(
                    test_case_ctx.test_case,
                    output,
                )
            else:
                ctx = contextvars.copy_context()
                hook_results = await global_state.event_loop().run_in_executor(
                    None,
                    ctx.run,
                    before_evaluators_hook,
                    test_case_ctx.test_case,
                    output,
                )

    # Revision usage is collected throughout a test case's run
    revision_usage = get_revision_usage()

    # Flush the logs before we send the result, since the CLI
    # accumulates the events and sends them as a batch along
    # with the result.
    flush()

    await post_to_cli(
        "/results",
        json=dict(
            testExternalId=test_id,
            runId=run_id,
            testCaseHash=test_case_ctx.hash(),
            testCaseBody=serialize_test_case(test_case_ctx.test_case),
            testCaseOutput=serialize(output),
            testCaseDurationMs=test_case_duration_ms,
            testCaseRevisionUsage=[usage.serialize() for usage in revision_usage] if revision_usage else None,
            testCaseHumanReviewInputFields=serialize_test_case_for_human_review(test_case_ctx.test_case),
            testCaseHumanReviewOutputFields=serialize_output_for_human_review(output),
        ),
    )
    return output, hook_results


async def run_test_case(
    test_id: str,
    run_id: Optional[str],
    test_case_ctx: TestCaseContext[TestCaseType],
    evaluators: Sequence[BaseTestEvaluator],
    fn: Union[Callable[[TestCaseType], Any], Callable[[TestCaseType], Awaitable[Any]]],
    before_evaluators_hook: Optional[Callable[[TestCaseType, Any], Any]],
) -> None:
    reset_token = test_case_run_context_var.set(
        TestCaseRunContext(
            run_id=run_id,
            test_id=test_id,
            test_case_hash=test_case_ctx.hash(),
        ),
    )
    try:
        output, hook_results = await run_test_case_unsafe(
            test_id=test_id,
            run_id=run_id,
            test_case_ctx=test_case_ctx,
            fn=fn,
            before_evaluators_hook=before_evaluators_hook,
        )
    except Exception as err:
        await send_error(
            test_id=test_id,
            run_id=run_id,
            test_case_hash=test_case_ctx.hash(),
            evaluator_id=None,
            error=err,
        )
        return

    try:
        await all_settled(
            [
                run_evaluator(
                    test_id=test_id,
                    run_id=run_id,
                    test_case_ctx=test_case_ctx,
                    output=output,
                    hook_results=hook_results,
                    evaluator=evaluator,
                )
                for evaluator in evaluators
            ],
        )
    except Exception as err:
        await send_error(
            test_id=test_id,
            run_id=run_id,
            test_case_hash=test_case_ctx.hash(),
            evaluator_id=None,
            error=err,
        )
    finally:
        test_case_run_context_var.reset(reset_token)


def validate_test_suite_inputs(
    test_id: str,
    test_cases: Sequence[TestCaseType],
    evaluators: Sequence[BaseTestEvaluator],
    grid_search_params: Optional[GridSearchParams],
) -> None:
    assert test_cases, f"[{test_id}] No test cases provided."
    for test_case in test_cases:
        assert isinstance(
            test_case,
            BaseTestCase,
        ), f"[{test_id}] Test case {test_case} does not implement {BaseTestCase.__name__}."
    for evaluator in evaluators:
        assert isinstance(
            evaluator,
            BaseTestEvaluator,
        ), f"[{test_id}] Evaluator {evaluator} does not implement {BaseTestEvaluator.__name__}."
    if grid_search_params is not None:
        assert isinstance(
            grid_search_params,
            dict,
        ), f"[{test_id}] grid_search_params must be a dict."
        for key, values in grid_search_params.items():
            assert isinstance(
                key,
                str,
            ), f"[{test_id}] grid_search_params keys must be strings."
            assert isinstance(
                values,
                Sequence,
            ), f"[{test_id}] grid_search_params values must be sequences."
            assert values, f"[{test_id}] grid_search_params values must not be empty."


async def send_info_for_alignment_mode(
    test_id: str,
    test_cases: Sequence[TestCaseType],
    caller_filepath: Optional[str],
) -> None:
    """
    Notifies the CLI with metadata about this test suite when running in "alignment mode",
    i.e. npx autoblocks testing align -- <cmd>
    """
    # Double check this is the correct test suite
    assert AutoblocksEnvVar.ALIGN_TEST_EXTERNAL_ID.get() == test_id

    # Tells the CLI what it needs to know about this test suite
    await post_to_cli(
        "/info",
        json=dict(
            language="python",
            runTestSuiteCalledFromDirectory=os.path.dirname(caller_filepath) if caller_filepath else None,
            testCaseHashes=[test_case.hash() for test_case in test_cases],
        ),
    )


def filter_test_cases_for_alignment_mode(
    test_id: str,
    test_cases: Sequence[TestCaseType],
) -> Sequence[TestCaseType]:
    """
    A test suite is in "alignment mode" if the user has started an "alignment session" via the CLI:

    $ npx autoblocks testing align --test-suite-id <test-suite-id> -- <cmd>

    This starts an interactive CLI where single test cases from the suite are run through `fn`,
    and the user then provides feedback on the output. This function checks for the environment
    variables that the CLI sets during an alignment session and filters the test cases accordingly.
    """
    # Double check this is the correct test suite
    assert AutoblocksEnvVar.ALIGN_TEST_EXTERNAL_ID.get() == test_id

    align_test_case_hash = AutoblocksEnvVar.ALIGN_TEST_CASE_HASH.get()
    if not align_test_case_hash:
        # The first time a test suite is run in alignment mode, the CLI will not provide a test case hash,
        # so we run the first one. On subsequent runs, the CLI will provide a test case hash since it
        # will have then received the /info request with the list of test case hashes.
        return [test_cases[0]]

    # Only run the selected test case
    return [test_case for test_case in test_cases if test_case.hash() == align_test_case_hash]


async def run_test_suite_for_grid_combo(
    test_id: str,
    test_cases: Sequence[TestCaseType],
    evaluators: Sequence[BaseTestEvaluator],
    fn: Union[Callable[[TestCaseType], Any], Callable[[TestCaseType], Awaitable[Any]]],
    before_evaluators_hook: Optional[Callable[[TestCaseType, Any], Any]],
    grid_search_run_group_id: Optional[str],
    grid_search_params_combo: Optional[GridSearchParamsCombo],
) -> None:
    start_resp = await post_to_cli(
        "/start",
        json=dict(
            testExternalId=test_id,
            gridSearchRunGroupId=grid_search_run_group_id,
            gridSearchParamsCombo=grid_search_params_combo,
        ),
    )

    run_id = None
    if start_resp:
        try:
            start_resp.raise_for_status()
        except HTTPStatusError:
            # Don't allow the run to continue if /start failed, since all subsequent
            # requests will fail if the CLI was not able to start the run.
            # Also note we don't need to send_error here, since the CLI will
            # have reported the HTTP error itself.
            return

        try:
            run_id = start_resp.json()["id"]
        except Exception:
            # We can drop this try-catch once everyone has updated their CLI
            # to the version that returns the run ID in the /start response.
            pass

    reset_token = grid_search_context_var.set(grid_search_params_combo) if grid_search_params_combo else None

    try:
        await all_settled(
            [
                run_test_case(
                    test_id=test_id,
                    run_id=run_id,
                    test_case_ctx=test_case_ctx,
                    evaluators=evaluators,
                    fn=fn,
                    before_evaluators_hook=before_evaluators_hook,
                )
                for test_case_ctx in yield_test_case_contexts_from_test_cases(test_cases)
            ],
        )
    except Exception as err:
        await send_error(
            test_id=test_id,
            run_id=run_id,
            test_case_hash=None,
            evaluator_id=None,
            error=err,
        )
    finally:
        if reset_token:
            grid_search_context_var.reset(reset_token)

    await post_to_cli(
        "/end",
        json=dict(testExternalId=test_id, runId=run_id),
    )


async def async_run_test_suite(
    test_id: str,
    test_cases: Sequence[TestCaseType],
    evaluators: Sequence[BaseTestEvaluator],
    fn: Union[Callable[[TestCaseType], Any], Callable[[TestCaseType], Awaitable[Any]]],
    before_evaluators_hook: Optional[Callable[[TestCaseType, Any], Any]],
    max_test_case_concurrency: int,
    caller_filepath: Optional[str],
    grid_search_params: Optional[GridSearchParams],
) -> None:
    if not AutoblocksEnvVar.CLI_SERVER_ADDRESS.get():
        log.warning(
            "\nRunning in debug mode since your tests are not being run within the context of the testing CLI; "
            "results will not be sent to Autoblocks.\n"
            "Make sure you are running your test command with:\n"
            "$ npx autoblocks testing exec -- <your test command>\n"
        )
    # Handle alignment mode
    align_test_id = AutoblocksEnvVar.ALIGN_TEST_EXTERNAL_ID.get()
    if align_test_id:
        if align_test_id != test_id:
            # Not the test suite in alignment mode
            return

        await send_info_for_alignment_mode(
            test_id=test_id,
            test_cases=test_cases,
            caller_filepath=caller_filepath,
        )
        test_cases = filter_test_cases_for_alignment_mode(
            test_id=test_id,
            test_cases=test_cases,
        )

    # This will be set if a user has triggered a run from the UI for a particular test suite.
    # If it is not this test suite, then we skip it.
    tests_and_hashes = tests_and_hashes_overrides_map()
    if tests_and_hashes:
        if test_id not in tests_and_hashes:
            log.info(f"Skipping test suite '{test_id}' because it is not in the list of test suites to run.")
            return

        # If the value for this test suite is non-empty, then it is a list of test case
        # hashes to run. We filter the test cases to only run those.
        hashes_to_run = set(tests_and_hashes[test_id] or [])
        if hashes_to_run:
            test_cases = [tc for tc in test_cases if tc.hash() in hashes_to_run]

    try:
        validate_test_suite_inputs(
            test_id=test_id,
            test_cases=test_cases,
            evaluators=evaluators,
            grid_search_params=grid_search_params,
        )
    except Exception as err:
        await send_error(
            test_id=test_id,
            run_id=None,
            test_case_hash=None,
            evaluator_id=None,
            error=err,
        )
        return

    # Initialize the semaphore registries
    test_case_semaphore_registry[test_id] = asyncio.Semaphore(max_test_case_concurrency)
    evaluator_semaphore_registry[test_id] = {
        evaluator.id: asyncio.Semaphore(evaluator.max_concurrency) for evaluator in evaluators
    }

    if grid_search_params is None:
        try:
            await run_test_suite_for_grid_combo(
                test_id=test_id,
                test_cases=test_cases,
                evaluators=evaluators,
                fn=fn,
                before_evaluators_hook=before_evaluators_hook,
                grid_search_run_group_id=None,
                grid_search_params_combo=None,
            )
        except Exception as err:
            await send_error(
                test_id=test_id,
                run_id=None,
                test_case_hash=None,
                evaluator_id=None,
                error=err,
            )
        return

    grid_resp = await post_to_cli(
        "/grids",
        json=dict(testExternalId=test_id, gridSearchParams=grid_search_params),
    )

    grid_search_run_group_id = None
    if grid_resp:
        try:
            grid_resp.raise_for_status()
        except HTTPStatusError:
            # Don't allow the run to continue if /grid failed, since all subsequent
            # requests will fail if the CLI was not able to create the grid.
            # Also note we don't need to send_error here, since the CLI will
            # have reported the HTTP error itself.
            return

        grid_search_run_group_id = grid_resp.json()["id"]

    try:
        await all_settled(
            [
                run_test_suite_for_grid_combo(
                    test_id=test_id,
                    test_cases=test_cases,
                    evaluators=evaluators,
                    fn=fn,
                    before_evaluators_hook=before_evaluators_hook,
                    grid_search_run_group_id=grid_search_run_group_id,
                    grid_search_params_combo=grid_params_combo,
                )
                for grid_params_combo in yield_grid_search_param_combos(grid_search_params)
            ],
        )
    except Exception as err:
        await send_error(
            test_id=test_id,
            run_id=None,
            test_case_hash=None,
            evaluator_id=None,
            error=err,
        )


# Sync fn
@overload
def run_test_suite(
    id: str,
    test_cases: Sequence[TestCaseType],
    fn: Callable[[TestCaseType], Any],
    evaluators: Optional[Sequence[BaseTestEvaluator]] = None,
    max_test_case_concurrency: int = DEFAULT_MAX_TEST_CASE_CONCURRENCY,
    before_evaluators_hook: Optional[Callable[[TestCaseType, Any], Any]] = None,
    grid_search_params: Optional[GridSearchParams] = None,
) -> None: ...


# Async fn
@overload
def run_test_suite(
    id: str,
    test_cases: Sequence[TestCaseType],
    fn: Callable[[TestCaseType], Awaitable[Any]],
    evaluators: Optional[Sequence[BaseTestEvaluator]] = None,
    max_test_case_concurrency: int = DEFAULT_MAX_TEST_CASE_CONCURRENCY,
    before_evaluators_hook: Optional[Callable[[TestCaseType, Any], Any]] = None,
    grid_search_params: Optional[GridSearchParams] = None,
) -> None: ...


def run_test_suite(
    id: str,
    test_cases: Sequence[TestCaseType],
    fn: Union[Callable[[TestCaseType], Any], Callable[[TestCaseType], Awaitable[Any]]],
    evaluators: Optional[Sequence[BaseTestEvaluator]] = None,
    # How many test cases to run concurrently
    max_test_case_concurrency: int = DEFAULT_MAX_TEST_CASE_CONCURRENCY,
    before_evaluators_hook: Optional[Callable[[TestCaseType, Any], Any]] = None,
    grid_search_params: Optional[GridSearchParams] = None,
) -> None:
    global_state.init()

    # Get the caller's filepath. Used in alignment mode to know where the test suite is located.
    try:
        caller_filepath = inspect.stack()[1].filename
    except Exception:
        caller_filepath = None

    asyncio.run_coroutine_threadsafe(
        async_run_test_suite(
            test_id=id,
            test_cases=test_cases,
            evaluators=evaluators or [],
            fn=fn,
            before_evaluators_hook=before_evaluators_hook,
            max_test_case_concurrency=max_test_case_concurrency,
            caller_filepath=caller_filepath,
            grid_search_params=grid_search_params,
        ),
        global_state.event_loop(),
    ).result()



import re
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import httpx

from autoblocks._impl.config.constants import API_ENDPOINT
from autoblocks._impl.config.constants import REVISION_LATEST
from autoblocks._impl.config.constants import REVISION_UNDEPLOYED
from autoblocks._impl.prompts.models import AutogeneratePromptsConfig
from autoblocks._impl.prompts.models import FrozenModel
from autoblocks._impl.prompts.placeholders import TemplatePlaceholder
from autoblocks._impl.prompts.placeholders import parse_placeholders_from_template
from autoblocks._impl.util import AutoblocksEnvVar


class Template(FrozenModel):
    id: str
    placeholders: List[TemplatePlaceholder]

    @property
    def snake_case_id(self) -> str:
        return to_snake_case(self.id)


class PromptCodegen(FrozenModel):
    id: str
    major_version: str
    params: Dict[str, Any]
    templates: List[Template]

    @property
    def title_case_id(self) -> str:
        return to_title_case(self.id)

    @property
    def params_class_name(self) -> str:
        return f"{self.title_case_id}Params"

    @property
    def template_renderer_class_name(self) -> str:
        return f"{self.title_case_id}TemplateRenderer"

    @property
    def execution_context_class_name(self) -> str:
        return f"{self.title_case_id}ExecutionContext"

    @property
    def manager_class_name(self) -> str:
        return f"{self.title_case_id}PromptManager"


HEADER = """############################################################################
# This file was generated automatically by Autoblocks. Do not edit directly.
############################################################################

from typing import Union  # noqa: F401

import pydantic  # noqa: F401

from autoblocks.prompts.context import PromptExecutionContext
from autoblocks.prompts.manager import AutoblocksPromptManager
from autoblocks.prompts.models import FrozenModel
from autoblocks.prompts.renderer import TemplateRenderer

"""


def to_title_case(s: str) -> str:
    # Remove leading numbers
    s = re.sub(r"^\d+", "", s)
    # Replace all non-alphanumeric characters with underscores
    s = re.sub(r"[^a-zA-Z0-9]+", "_", s)
    # Replace all underscores with capital letter of next word
    return re.sub(r"(?:^|_)(.)", lambda m: m.group(1).upper(), s)


def to_snake_case(s: str) -> str:
    # Remove leading numbers
    s = re.sub(r"^\d+", "", s)

    # Replace all non-alphanumeric characters with spaces
    s = re.sub(r"[^a-zA-Z0-9]+", " ", s)

    # Replace spaces with underscores
    s = re.sub(r"\s+", "_", s)

    # Add underscores between camel case words
    s = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", s)
    s = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", s)

    # Remove leading and trailing underscores
    s = s.strip("_")

    return s.lower()


def infer_type(value: Any) -> Optional[str]:
    if isinstance(value, str):
        return "str"
    elif isinstance(value, bool):
        return "bool"
    elif isinstance(value, (int, float)):
        # The default union mode in pydantic is "smart" mode,
        # which will use the most specific type possible.
        # https://docs.pydantic.dev/latest/concepts/unions/
        return "Union[float, int]"
    elif isinstance(value, list):
        if len(value) > 0:
            return f"list[{infer_type(value[0])}]"
    return None


def indent(times: int = 1, size: int = 4) -> str:
    return " " * size * times


def generate_params_class_code(prompt: PromptCodegen) -> str:
    auto = f"class {prompt.params_class_name}(FrozenModel):\n"
    if not prompt.params:
        auto += f"{indent()}pass\n"
        return auto

    for key, value in prompt.params.items():
        type_hint = infer_type(value)
        if type_hint is None:
            continue
        snake_case_key = to_snake_case(key)
        auto += f'{indent()}{snake_case_key}: {type_hint} = pydantic.Field(..., alias="{key}")\n'

    return auto


def generate_template_render_method_code(template: Template) -> str:
    auto = f"{indent()}def {template.snake_case_id}(\n{indent(2)}self,\n"

    if template.placeholders:
        # Require all params to be passed in as keyword arguments
        auto += f"{indent(2)}*,\n"

    for placeholder in template.placeholders:
        auto += f"{indent(2)}{to_snake_case(placeholder.name)}: str,\n"

    auto += f"{indent()}) -> str:\n"
    auto += f'{indent(2)}return self._render(\n{indent(3)}"{template.id}",\n'

    for placeholder in template.placeholders:
        kwarg_name = to_snake_case(placeholder.name)
        auto += f"{indent(3)}{kwarg_name}={kwarg_name},\n"

    auto += f"{indent(2)})\n"

    return auto


def generate_template_renderer_class_code(prompt: PromptCodegen) -> str:
    auto = f"class {prompt.template_renderer_class_name}(TemplateRenderer):\n"

    # Add name mapper class attribute
    # The name mapper maps the original template placeholder name
    # to the snake case name of the corresponding keyword argument
    name_mapper = {}
    for template in prompt.templates:
        for placeholder in template.placeholders:
            # We should have filtered out escaped placeholders earlier
            assert not placeholder.is_escaped
            name_mapper[placeholder.name] = to_snake_case(placeholder.name)

    if name_mapper:
        auto += f"{indent()}__name_mapper__ = {{\n"

        for key in sorted(name_mapper.keys()):
            auto += f'{indent(2)}"{key}": "{name_mapper[key]}",\n'

        auto += f"{indent()}}}\n\n"
    else:
        auto += f"{indent()}__name_mapper__ = {{}}\n\n"

    auto += "\n".join(generate_template_render_method_code(template) for template in prompt.templates)

    return auto


def generate_execution_context_class_code(prompt: PromptCodegen) -> str:
    auto = f"class {prompt.execution_context_class_name}(\n"
    auto += f"{indent()}PromptExecutionContext[\n"
    auto += f"{indent(2)}{prompt.params_class_name},\n"
    auto += f"{indent(2)}{prompt.template_renderer_class_name},\n"
    auto += f"{indent()}],\n"
    auto += "):\n"
    auto += f"{indent()}__params_class__ = {prompt.params_class_name}\n"
    auto += f"{indent()}__template_renderer_class__ = {prompt.template_renderer_class_name}\n"

    return auto


def generate_prompt_manager_class_code(prompt: PromptCodegen) -> str:
    auto = f"class {prompt.manager_class_name}(\n"
    auto += f"{indent()}AutoblocksPromptManager[{prompt.execution_context_class_name}],\n"
    auto += "):\n"
    auto += f'{indent()}__prompt_id__ = "{prompt.id}"\n'
    auto += f'{indent()}__prompt_major_version__ = "{prompt.major_version}"\n'
    auto += f"{indent()}__execution_context_class__ = {prompt.execution_context_class_name}\n"
    return auto + "\n"


def generate_code_for_prompt(prompt: PromptCodegen) -> str:
    return "\n\n".join(
        [
            x
            for x in [
                generate_params_class_code(prompt),
                generate_template_renderer_class_code(prompt),
                generate_execution_context_class_code(prompt),
                generate_prompt_manager_class_code(prompt),
            ]
            if x is not None
        ],
    )


def make_prompts_from_config(
    config: AutogeneratePromptsConfig,
) -> List[PromptCodegen]:
    """
    Fetches prompt data from the Autoblocks API and generates code for each prompt
    configured in the .autoblocks.yml config file.
    """
    prompts = []

    for prompt in config.prompts:
        if prompt.major_version is not None:
            major = prompt.major_version
            minor = REVISION_LATEST
        elif prompt.dangerously_use_undeployed_revision is not None:
            major = REVISION_UNDEPLOYED
            minor = prompt.dangerously_use_undeployed_revision
        else:
            raise ValueError(
                f"Error in {prompt.id} config: "
                f"Either `major_version` or `dangerously_use_undeployed_revision` must be specified"
            )

        resp = httpx.get(
            url=f"{API_ENDPOINT}/prompts/{prompt.id}/major/{major}/minor/{minor}",
            headers={"Authorization": f"Bearer {AutoblocksEnvVar.API_KEY.get()}"},
        )
        resp.raise_for_status()
        data = resp.json()

        try:
            params = data["params"]["params"] or {}
        except (KeyError, TypeError):
            params = {}

        templates = []
        for template in data["templates"]:
            template_id = template["id"]
            template_placeholders = parse_placeholders_from_template(template["template"])
            templates.append(
                Template(
                    id=template_id,
                    placeholders=[p for p in template_placeholders if not p.is_escaped],
                ),
            )

        prompts.append(
            PromptCodegen(
                id=prompt.id,
                major_version=major,
                params=params,
                templates=sorted(templates, key=lambda t: t.snake_case_id),
            )
        )

    return sorted(prompts, key=lambda p: p.title_case_id)


def generate_code_for_config(config: AutogeneratePromptsConfig) -> str:
    prompts = make_prompts_from_config(config)

    auto = [HEADER]

    for prompt in prompts:
        auto.append(generate_code_for_prompt(prompt))

    return "\n".join(auto).strip() + "\n"


def write_generated_code_for_config(config: AutogeneratePromptsConfig) -> None:
    """Write generated code to outfile."""
    with open(config.outfile, "w") as f:
        f.write(generate_code_for_config(config))



import abc
import json
from textwrap import dedent
from typing import Any
from typing import Generic
from typing import List
from typing import Optional

from autoblocks._impl import global_state
from autoblocks._impl.config.constants import API_ENDPOINT
from autoblocks._impl.testing.evaluators.util import get_autoblocks_api_key
from autoblocks._impl.testing.evaluators.util import get_openai_client
from autoblocks._impl.testing.evaluators.util import get_test_id
from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import EvaluationOverride
from autoblocks._impl.testing.models import EvaluationOverrideComment
from autoblocks._impl.testing.models import EvaluationOverrideField
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import ScoreChoice
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.models import Threshold
from autoblocks._impl.util import encode_uri_component

function_name = "select_answer"


class BaseLLMJudge(BaseTestEvaluator, abc.ABC, Generic[TestCaseType, OutputType]):
    """
    Base evaluator for creating an LLM judge.
    """

    @property
    def threshold(self) -> Optional[Threshold]:
        """
        The threshold for the evaluator.
        """
        return None

    @property
    def model(self) -> str:
        """
        The model to use for the evaluator.
        It must be an OpenAI model that supports tools.
        Defaults to "gpt-4-turbo".
        """
        return "gpt-4-turbo"

    @property
    def num_overrides(self) -> int:
        """
        The number of recent evaluation overrides to fetch for the evaluator and pass to make_prompt.

        Defaults to 0.
        """
        return 0

    @property
    @abc.abstractmethod
    def score_choices(self) -> List[ScoreChoice]:
        """
        The choices for the LLM judge to use when answering.
        """
        pass

    @abc.abstractmethod
    def make_prompt(
        self, test_case: TestCaseType, output: OutputType, recent_overrides: List[EvaluationOverride]
    ) -> str:
        """
        The prompt passed to the LLM judge. Should be poised as a question.
        """
        pass

    def _find_score_choice_from_value(self, value: float) -> Optional[ScoreChoice]:
        """
        Find the score choice from choices based the float value.
        """
        return next((score for score in self.score_choices if score.value == value), None)

    def _make_recent_overrides_url(self) -> str:
        test_id = get_test_id(evaluator_id=self.id)
        encoded_test_id = encode_uri_component(test_id)
        encoded_evaluator_id = encode_uri_component(self.id)
        url = f"{API_ENDPOINT}/test-suites/{encoded_test_id}/evaluators/{encoded_evaluator_id}/human-reviews"
        return f"{url}?n={self.num_overrides}"

    async def _get_recent_overrides(self) -> List[EvaluationOverride]:
        if self.num_overrides == 0:
            # Don't fetch if the consumer hasn't requested any overrides
            return []

        resp = await global_state.http_client().get(
            self._make_recent_overrides_url(),
            headers={"Authorization": f"Bearer {get_autoblocks_api_key(self.id)}"},
        )
        resp.raise_for_status()
        data = resp.json()
        overrides = []
        for eval_data in data:
            override_score = self._find_score_choice_from_value(eval_data["overrideScore"])
            original_score = self._find_score_choice_from_value(eval_data["originalScore"])
            if override_score is None or original_score is None:
                # If for some reason we can't find the score choice, skip this override
                # Could happen if the score choices have changed since the override was created
                continue

            overrides.append(
                EvaluationOverride(
                    original_score=original_score,
                    override_score=override_score,
                    output_fields=[EvaluationOverrideField(**field) for field in eval_data["outputFields"]],
                    input_fields=[EvaluationOverrideField(**field) for field in eval_data["inputFields"]],
                    comments=[
                        EvaluationOverrideComment(
                            field_id=comment["fieldId"],
                            quoted_text=comment["quotedText"],
                            comment_text=comment["commentText"],
                        )
                        for comment in eval_data["comments"]
                    ],
                )
            )

        return overrides

    def _make_tool(self) -> dict[str, Any]:
        """
        We use function calling to force the LLM to return structured JSON.

        See https://platform.openai.com/docs/guides/function-calling
        """
        return {
            "type": "function",
            "function": {
                "name": function_name,
                "description": "Call this function to select an answer",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "reason": {
                            "type": "string",
                            "description": "The reason for the answer",
                        },
                        "answer": {
                            "type": "string",
                            "description": "The answer to select",
                            "enum": [score.name for score in self.score_choices],
                        },
                    },
                    "required": ["reason", "answer"],
                },
            },
        }

    async def _execute_prompt(self, test_case: TestCaseType, output: OutputType) -> Evaluation:
        recent_overrides = await self._get_recent_overrides()
        prompt = self.make_prompt(test_case=test_case, output=output, recent_overrides=recent_overrides)
        response = await get_openai_client(evaluator_id=self.id).chat.completions.create(
            model=self.model,
            temperature=0.0,
            tool_choice={"type": "function", "function": {"name": function_name}},
            tools=[self._make_tool()],
            messages=[
                dict(
                    role="system",
                    content=dedent(
                        """Answer the following question by selecting an answer.
                            Always provide a reason for your answer."""
                    ),
                ),
                dict(role="user", content=prompt),
            ],
        )
        tool_call = response.choices[0].message.tool_calls[0]
        if tool_call.function.name != function_name:
            # This shouldn't happen since we force the LLM to call our function, but we check just in case
            raise ValueError(f"Unexpected tool call: {tool_call}")

        function_args = json.loads(tool_call.function.arguments)

        # We should always be able to find the score choice since we set the scores as an enum in the tool
        score = next((score for score in self.score_choices if score.name == function_args["answer"]), None)
        if score is None:
            raise ValueError(f"Unexpected answer: {function_args['answer']}")

        return Evaluation(
            score=score.value,
            threshold=self.threshold,
            metadata={
                "reason": function_args["reason"],
                "prompt": prompt,
            },
        )

    async def evaluate_test_case(self, test_case: TestCaseType, output: OutputType) -> Evaluation:
        return await self._execute_prompt(test_case=test_case, output=output)



import abc
from typing import Generic
from typing import List

from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.models import Threshold


class BaseHasAllSubstrings(BaseTestEvaluator, abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The HasAllSubstrings evaluator checks if the output contains all the expected substrings.
    Scores 1 if all substrings are present, 0 otherwise.
    The comparison is case-sensitive.
    """

    @abc.abstractmethod
    def test_case_mapper(self, test_case: TestCaseType) -> List[str]:
        """
        Map your test case to a list of strings to check for in the output.
        """
        pass

    @abc.abstractmethod
    def output_mapper(self, output: OutputType) -> str:
        """
        Map your output to a string for comparison.
        """
        pass

    def evaluate_test_case(self, test_case: TestCaseType, output: OutputType) -> Evaluation:
        expected_substrings = self.test_case_mapper(test_case)
        mapped_output = self.output_mapper(output)
        missing_substrings = [s for s in expected_substrings if s not in mapped_output]
        score = 0 if missing_substrings else 1
        return Evaluation(score=score, threshold=Threshold(gte=1), metadata={"missing_substrings": missing_substrings})



import asyncio
import contextvars
import dataclasses
import inspect
import logging
import uuid
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from autoblocks._impl import global_state
from autoblocks._impl.config.constants import INGESTION_ENDPOINT
from autoblocks._impl.context_vars import test_case_run_context_var
from autoblocks._impl.testing.models import BaseEventEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import HumanReviewField
from autoblocks._impl.testing.models import TracerEvent
from autoblocks._impl.testing.util import serialize_human_review_fields
from autoblocks._impl.util import AnyTask
from autoblocks._impl.util import AutoblocksEnvVar
from autoblocks._impl.util import all_settled
from autoblocks._impl.util import get_running_loop
from autoblocks._impl.util import now_iso_8601

log = logging.getLogger(__name__)

evaluator_semaphore_registry: dict[str, asyncio.Semaphore] = {}  # evaluator_id -> semaphore


@dataclasses.dataclass(frozen=True)
class Payload:
    message: str
    trace_id: Optional[str]
    timestamp: str
    properties: dict[str, Any]
    system_properties: Optional[dict[str, Any]]

    def to_json(self) -> dict[str, Any]:
        return {
            "message": self.message,
            "traceId": self.trace_id,
            "timestamp": self.timestamp,
            "properties": self.properties,
            "systemProperties": self.system_properties or None,
        }


class AutoblocksTracer:
    def __init__(
        self,
        ingestion_key: Optional[str] = None,
        # Initialize the tracer with a trace_id. All events sent with this tracer will
        # send this trace ID unless overwritten by:
        # - calling set_trace_id
        # - specifying a trace_id when calling send_event
        trace_id: Optional[str] = None,
        # Initialize the tracer with properties. All events sent with this tracer will
        # send these properties unless overwritten by:
        # - calling set_properties
        # Additionally, these properties can be updated by:
        # - calling update_properties
        # - specifying properties when calling send_event
        properties: Optional[Dict[Any, Any]] = None,
        # Timeout for sending events to Autoblocks
        timeout: timedelta = timedelta(seconds=5),
    ):
        global_state.init()  # Start up event loop if not already started
        self._trace_id: Optional[str] = trace_id
        self._properties: Dict[Any, Any] = properties or {}

        ingestion_key = ingestion_key or AutoblocksEnvVar.INGESTION_KEY.get()
        if not ingestion_key:
            raise ValueError(
                f"You must provide an ingestion_key or set the {AutoblocksEnvVar.INGESTION_KEY} environment variable."
            )

        self._client_headers = {"Authorization": f"Bearer {ingestion_key}"}
        self._timeout_seconds = timeout.total_seconds()
        self._cli_server_address = AutoblocksEnvVar.CLI_SERVER_ADDRESS.get()
        self._throw_on_error = AutoblocksEnvVar.TRACER_THROW_ON_ERROR.get() == "1"

    def set_trace_id(self, trace_id: str) -> None:
        """
        Set the trace ID for all events sent by this tracer.
        """
        self._trace_id = trace_id

    @property
    def trace_id(self) -> Optional[str]:
        """
        Get the trace ID currently set on the tracer.
        """
        return self._trace_id

    def set_properties(self, properties: Dict[Any, Any]) -> None:
        """
        Set the properties for all events sent by this tracer.

        NOTE: This will overwrite any existing properties.
        """
        self._properties = properties

    def update_properties(self, properties: Dict[Any, Any]) -> None:
        """
        Update the properties for all events sent by this tracer.
        """
        self._properties.update(properties)

    @staticmethod
    def _evaluation_to_json(evaluation: Evaluation, evaluator_external_id: str) -> Dict[str, Any]:
        return dict(
            id=str(uuid.uuid4()),
            score=evaluation.score,
            threshold=dataclasses.asdict(evaluation.threshold) if evaluation.threshold else None,
            metadata=evaluation.metadata,
            evaluatorExternalId=evaluator_external_id,
        )

    @staticmethod
    async def _evaluate_event_unsafe(
        event: TracerEvent,
        evaluator: BaseEventEvaluator,
    ) -> Optional[Evaluation]:
        """
        This function is suffixed with _unsafe because it doesn't handle exceptions.
        Its caller will catch and handle all exceptions.
        """
        if evaluator.id not in evaluator_semaphore_registry:
            evaluator_semaphore_registry[evaluator.id] = asyncio.Semaphore(evaluator.max_concurrency)

        async with evaluator_semaphore_registry[evaluator.id]:
            if inspect.iscoroutinefunction(evaluator.evaluate_event):
                # mypy error:
                # error: Returning Any from function declared to return "Evaluation | None"  [no-any-return]
                # Not sure what mypy wants here, evaluate_event is typed correctly
                return await evaluator.evaluate_event(event)  # type: ignore
            else:
                ctx = contextvars.copy_context()
                return await global_state.event_loop().run_in_executor(
                    None,
                    ctx.run,  # type: ignore
                    evaluator.evaluate_event,
                    event,
                )

    async def _evaluate_event(self, event: TracerEvent, evaluator: BaseEventEvaluator) -> Optional[Evaluation]:
        """
        Evaluates an event using a provided evaluator.
        """
        try:
            return await self._evaluate_event_unsafe(
                event=event,
                evaluator=evaluator,
            )
        except Exception as err:
            log.error(f"Unable to execute evaluator with id: {evaluator.id}. Error: %s", err, exc_info=True)
            return None

    async def _run_evaluators_unsafe(
        self,
        event: TracerEvent,
        evaluators: Sequence[BaseEventEvaluator],
    ) -> List[Dict[str, Any]]:
        """
        Run a list of evaluators on an event and return the evaluations as a list of dictionaries.

        This is suffixed with _unsafe because it doesn't handle exceptions.
        Its caller will catch and handle all exceptions.
        """
        if len(evaluators) == 0:
            return []
        evaluations: Tuple[Union[Optional[Evaluation], BaseException]] = await all_settled(
            [
                self._evaluate_event(
                    event=event,
                    evaluator=evaluator,
                )
                for evaluator in evaluators
            ],
        )
        return [
            self._evaluation_to_json(evaluation=evaluation, evaluator_external_id=evaluator.id)
            for evaluator, evaluation in zip(evaluators, evaluations)
            if isinstance(evaluation, Evaluation)
        ]

    async def _async_send_event(
        self,
        payload: Payload,
        evaluators: Optional[Sequence[BaseEventEvaluator]],
    ) -> None:
        # If there are evaluators, run them and compute the evaluations property
        if evaluators:
            try:
                evaluations = await self._run_evaluators_unsafe(
                    event=TracerEvent(
                        message=payload.message,
                        trace_id=payload.trace_id,
                        timestamp=payload.timestamp,
                        properties=payload.properties,
                    ),
                    evaluators=evaluators,
                )
                if evaluations:
                    # Update merged properties with computed evaluations property
                    payload = dataclasses.replace(
                        payload,
                        properties={
                            **payload.properties,
                            "evaluations": evaluations,
                        },
                    )
            except Exception as err:
                log.error("Unable to evaluate events. Error: %s", err, exc_info=True)

        try:
            if global_state.main_thread_has_finished():
                # If we're in a shutdown state, we need to use the sync client
                # to avoid scheduling new futures after the interpreter has shut down.
                # See https://github.com/boto/boto3/issues/3113
                log.debug("Sending event with SYNC client because interpreter has shut down")
                req = global_state.sync_http_client().post(
                    url=INGESTION_ENDPOINT,
                    json=payload.to_json(),
                    headers=self._client_headers,
                    timeout=self._timeout_seconds,
                )
            else:
                log.debug("Sending event with ASYNC client")
                req = await global_state.http_client().post(
                    url=INGESTION_ENDPOINT,
                    json=payload.to_json(),
                    headers=self._client_headers,
                    timeout=self._timeout_seconds,
                )
            req.raise_for_status()
        except Exception as err:
            # Since we don't call task.result() in the caller of this function,
            # we need to catch and log errors here in order for them to be
            # seen.
            log.error(f"Failed to send event to Autoblocks: {err}", exc_info=True)
            if self._throw_on_error:
                raise err

    def _sync_send_test_event_unsafe(
        self,
        payload: Payload,
    ) -> None:
        """Sends an event to the Autoblocks CLI in a testing environment."""
        if not self._cli_server_address:
            log.error("Failed to send test event to Autoblocks. CLI server address not set.")
            return
        test_case_run = test_case_run_context_var.get()
        if not test_case_run:
            log.error("Failed to send test event to Autoblocks. No test ID set.")
            return

        req = global_state.sync_http_client().post(
            url=f"{self._cli_server_address}/events",
            json=dict(
                testExternalId=test_case_run.test_id,
                runId=test_case_run.run_id,
                testCaseHash=test_case_run.test_case_hash,
                event=payload.to_json(),
            ),
            timeout=self._timeout_seconds,
        )
        req.raise_for_status()

    def _make_request_payload(
        self,
        *,
        message: str,
        trace_id: Optional[str],
        timestamp: Optional[str],
        span_id: Optional[str],
        parent_span_id: Optional[str],
        properties: Optional[Dict[str, Any]],
        prompt_tracking: Optional[Dict[str, Any]],
        human_review_fields: Optional[List[HumanReviewField]],
    ) -> Payload:
        trace_id = trace_id or self._trace_id
        timestamp = timestamp or now_iso_8601()

        merged_properties = dict(self._properties)
        merged_properties.update(properties or {})

        # Overwrite properties with any provided as top level arguments
        if span_id:
            merged_properties["span_id"] = span_id
        if parent_span_id:
            merged_properties["parent_span_id"] = parent_span_id
        if prompt_tracking:
            merged_properties["promptTracking"] = prompt_tracking

        system_properties = {}
        if human_review_fields:
            system_properties["humanReviewFields"] = serialize_human_review_fields(human_review_fields)

        return Payload(
            message=message,
            trace_id=trace_id,
            timestamp=timestamp,
            properties=merged_properties,
            system_properties=system_properties,
        )

    def _dispatch_event_unsafe(
        self,
        payload: Payload,
        evaluators: Optional[Sequence[BaseEventEvaluator]],
    ) -> None:
        task: AnyTask

        # Check if we're already in a running event loop
        if running_loop := get_running_loop():
            # If we are, execute the task on that loop
            task = running_loop.create_task(
                self._async_send_event(
                    payload=payload,
                    evaluators=evaluators,
                ),
            )
        else:
            # Otherwise, send the task to our background loop
            task = asyncio.run_coroutine_threadsafe(
                self._async_send_event(
                    payload=payload,
                    evaluators=evaluators,
                ),
                global_state.event_loop(),
            )

        # Add the task to the set of tasks the
        # global state will wait to complete
        # on shutdown
        global_state.add_background_task(task)

        if self._throw_on_error:
            # Wait for the result if AUTOBLOCKS_TRACER_THROW_ON_ERROR is set to true,
            # since we set this environment variable in end-to-end testing contexts
            # where we want to raise exceptions immediately.
            task.result()

    def send_event(
        self,
        message: str,
        # Require all arguments after message to be specified via key=value
        *,
        trace_id: Optional[str] = None,
        span_id: Optional[str] = None,
        parent_span_id: Optional[str] = None,
        timestamp: Optional[str] = None,
        properties: Optional[Dict[str, Any]] = None,
        evaluators: Optional[Sequence[BaseEventEvaluator]] = None,
        prompt_tracking: Optional[Dict[str, Any]] = None,
        human_review_fields: Optional[List[HumanReviewField]] = None,
    ) -> None:
        """
        Sends an event to the Autoblocks ingestion API.
        """
        try:
            # Prepare request payload
            payload = self._make_request_payload(
                message=message,
                trace_id=trace_id,
                timestamp=timestamp,
                span_id=span_id,
                parent_span_id=parent_span_id,
                properties=properties,
                prompt_tracking=prompt_tracking,
                human_review_fields=human_review_fields,
            )

            # If the test case run contextvar is set, we are in a test context and should send events to the CLI.
            # We also do not run evaluators in a test context, so the only argument we pass here is `payload`.
            if test_case_run_context_var.get() is not None:
                self._sync_send_test_event_unsafe(payload)
                return

            # Otherwise we are in a real context and should send events to the Autoblocks ingestion API
            self._dispatch_event_unsafe(
                payload=payload,
                evaluators=evaluators,
            )
        except Exception as err:
            log.error(f"Failed to send event to Autoblocks: {err}", exc_info=True)
            if self._throw_on_error:
                raise err



import abc
from typing import Generic

from autoblocks._impl.testing.evaluators.ragas_base import BaseRagas
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType


class BaseRagasContextRelevancy(BaseRagas[TestCaseType, OutputType], abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The RagasContextRelevancy evaluator evaluates the relevancy of the retrieved context,
    calculated based on both the question and contexts

    See more: https://docs.ragas.io/en/stable/concepts/metrics/context_relevancy.html
    """

    metric_name = "context_relevancy"



from autoblocks._impl.testing.models import BaseEvaluator
from autoblocks._impl.testing.models import BaseEventEvaluator
from autoblocks._impl.testing.models import BaseTestCase
from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import EvaluationOverride
from autoblocks._impl.testing.models import EvaluationOverrideComment
from autoblocks._impl.testing.models import EvaluationOverrideField
from autoblocks._impl.testing.models import HumanReviewField
from autoblocks._impl.testing.models import ScoreChoice
from autoblocks._impl.testing.models import TestCaseConfig
from autoblocks._impl.testing.models import Threshold
from autoblocks._impl.testing.models import TracerEvent

__all__ = [
    "BaseEvaluator",
    "BaseEventEvaluator",
    "BaseTestCase",
    "BaseTestEvaluator",
    "Evaluation",
    "EvaluationOverride",
    "EvaluationOverrideComment",
    "EvaluationOverrideField",
    "HumanReviewField",
    "TestCaseConfig",
    "Threshold",
    "TracerEvent",
    "ScoreChoice",
]



import abc
import functools
from typing import Any
from typing import Dict
from typing import Generic
from typing import Type
from typing import TypeVar

import pydantic

from autoblocks._impl.prompts.models import Prompt
from autoblocks._impl.prompts.renderer import TemplateRenderer

ParamsType = TypeVar("ParamsType", bound=pydantic.BaseModel)
TemplateRendererType = TypeVar("TemplateRendererType", bound=TemplateRenderer)


class PromptExecutionContext(abc.ABC, Generic[ParamsType, TemplateRendererType]):
    __params_class__: Type[ParamsType]
    __template_renderer_class__: Type[TemplateRendererType]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        for attr in (
            "__params_class__",
            "__template_renderer_class__",
        ):
            if not hasattr(cls, attr):
                raise ValueError(f"PromptExecutionContext subclass {cls} must define {attr}")

    def __init__(self, prompt: Prompt) -> None:
        self._prompt = prompt

    @functools.cached_property
    def params(self) -> ParamsType:
        if self._prompt.params:
            params = self._prompt.params.params or {}
        else:
            params = {}
        return self.__params_class__(
            **params,
        )

    @functools.cached_property
    def render(self) -> TemplateRendererType:
        return self.__template_renderer_class__(
            self._prompt,
        )

    def track(self) -> Dict[str, Any]:
        return self._prompt.tracking



import abc
from typing import Generic

from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.models import Threshold


class BaseIsEquals(BaseTestEvaluator, abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The IsEquals evaluator checks if the output equals the expected output.
    The comparison is case-sensitive. Scores 1 if it is equal, 0 otherwise.
    """

    @abc.abstractmethod
    def test_case_mapper(self, test_case: TestCaseType) -> str:
        """
        Map your test case to a string for comparison.
        """
        pass

    @abc.abstractmethod
    def output_mapper(self, output: OutputType) -> str:
        """
        Map your output to a string for comparison.
        """
        pass

    def evaluate_test_case(self, test_case: TestCaseType, output: OutputType) -> Evaluation:
        actual_output = self.output_mapper(output)
        expected_output = self.test_case_mapper(test_case)
        score = 1 if actual_output == expected_output else 0
        metadata = {"expected_output": expected_output, "actual_output": actual_output} if score == 0 else None
        return Evaluation(
            score=score,
            threshold=Threshold(gte=1),
            metadata=metadata,
        )



import abc
import asyncio
import json
import logging
from datetime import timedelta
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generic
from typing import Optional
from typing import TypeVar

from autoblocks._impl import global_state
from autoblocks._impl.config.constants import API_ENDPOINT
from autoblocks._impl.config.constants import REVISION_LATEST
from autoblocks._impl.configs.models import RemoteConfig
from autoblocks._impl.configs.models import RemoteConfigResponse
from autoblocks._impl.context_vars import RevisionType
from autoblocks._impl.context_vars import register_revision_usage
from autoblocks._impl.util import AnyTask
from autoblocks._impl.util import AutoblocksEnvVar
from autoblocks._impl.util import get_running_loop

log = logging.getLogger(__name__)

AutoblocksConfigValueType = TypeVar("AutoblocksConfigValueType")


def is_testing_context() -> bool:
    """
    Note that we check for the presence of the CLI environment
    variable and not the test case contextvars because the
    contextvars aren't set until run_test_suite is called,
    whereas a config manager might have already been imported
    and initialized by the time run_test_suite is called.
    """
    return bool(AutoblocksEnvVar.CLI_SERVER_ADDRESS.get())


def config_revisions_map() -> dict[str, str]:
    """
    The AUTOBLOCKS_OVERRIDES_CONFIG_REVISIONS environment variable is a JSON-stringified
    map of config IDs to revision IDs. This is set in CI test runs triggered
    from the UI.
    """
    if not is_testing_context():
        return {}

    config_revisions_raw = AutoblocksEnvVar.OVERRIDES_CONFIG_REVISIONS.get()
    if not config_revisions_raw:
        return {}
    return json.loads(config_revisions_raw)  # type: ignore


def make_request_url(config: RemoteConfig) -> str:
    config_id = config.id
    base = f"{API_ENDPOINT}/configs/{config_id}"
    if not isinstance(config, RemoteConfig) and (
        (config.major_version is None and config.minor_version is None)
        or config.dangerously_use_undeployed_revision is None
    ):
        raise ValueError(f"Invalid config type: {config}")

    if is_testing_context() and (revision_id := config_revisions_map().get(config_id)):
        return f"{base}/revisions/{revision_id}"

    if config.dangerously_use_undeployed_revision is not None:
        if config.dangerously_use_undeployed_revision == REVISION_LATEST:
            return f"{base}/revisions/{REVISION_LATEST}"
        else:
            return f"{base}/revisions/{config.dangerously_use_undeployed_revision}"

    if config.minor_version == REVISION_LATEST:
        return f"{base}/major/{config.major_version}/minor/{REVISION_LATEST}"

    return f"{base}/major/{config.major_version}/minor/{config.minor_version}"


def is_remote_config_refreshable(config: RemoteConfig) -> bool:
    """
    A config is refreshable if the user has specified that they want the latest deployed config
    or the latest undeployed config.
    """
    return (
        config.major_version is not None and config.minor_version == REVISION_LATEST
    ) or config.dangerously_use_undeployed_revision == REVISION_LATEST


async def get_remote_config(config: RemoteConfig, timeout: timedelta, api_key: str) -> RemoteConfigResponse:
    resp = await global_state.http_client().get(
        make_request_url(config=config),
        timeout=timeout.total_seconds(),
        headers={"Authorization": f"Bearer {api_key}"},
    )
    resp.raise_for_status()
    return RemoteConfigResponse.model_validate(resp.json())


class AutoblocksConfig(
    abc.ABC,
    Generic[AutoblocksConfigValueType],
):
    _value: AutoblocksConfigValueType
    _remote_config_id: Optional[str] = None
    _remote_config_revision_id: Optional[str] = None

    def __init__(self, value: AutoblocksConfigValueType) -> None:
        self._value = value

    async def _refresh_loop(
        self,
        config: RemoteConfig,
        refresh_interval: timedelta,
        refresh_timeout: timedelta,
        api_key: str,
        parser: Callable[[Dict[str, Any]], AutoblocksConfigValueType],
    ) -> None:
        """
        Refreshes the latest version of the config every `refresh_interval` seconds.

        This isn't exactly "every n seconds" but more "run and then wait n seconds",
        which will cause it to drift each run by the time it takes to run the refresh,
        but I think that is ok in this scenario.
        """
        while True:
            await asyncio.sleep(refresh_interval.total_seconds())

            try:
                await self._load_and_set_remote_config(
                    config=config, api_key=api_key, timeout=refresh_timeout, parser=parser
                )
            except Exception as err:
                log.warning(f"Failed to refresh config '{config.id}': {err}")

    async def _load_and_set_remote_config(
        self,
        config: RemoteConfig,
        api_key: str,
        timeout: timedelta,
        parser: Callable[[Dict[str, Any]], AutoblocksConfigValueType],
    ) -> None:
        """
        Loads the remote config from Autoblocks and sets the value of this config
        based on the result.

        If the parser is specified, the value will be parsed using the parser.
        """
        remote_config = await get_remote_config(config, timeout=timeout, api_key=api_key)
        try:
            obj = {}
            for prop in remote_config.properties:
                obj[prop.id] = prop.value
            parsed = parser(obj)
            if parsed is not None:
                self._value = parsed
                self._remote_config_id = remote_config.id
                self._remote_config_revision_id = remote_config.revision_id
        except Exception as err:
            log.error(f"Failed to parse config '{config.id}': {err}", exc_info=True)
            raise err

    def _activate_from_remote_unsafe(
        self,
        config: RemoteConfig,
        refresh_interval: timedelta,
        refresh_timeout: timedelta,
        activate_timeout: timedelta,
        parser: Callable[[Dict[str, Any]], AutoblocksConfigValueType],
        api_key: Optional[str] = None,
    ) -> None:
        api_key = api_key or AutoblocksEnvVar.API_KEY.get()
        if not api_key:
            raise ValueError(
                f"You must either pass in the API key via 'api_key' or "
                f"set the {AutoblocksEnvVar.API_KEY} environment variable."
            )

        global_state.init()
        task: AnyTask
        if running_loop := get_running_loop():
            # If we're already in a running loop, execute the task on that loop
            task = running_loop.create_task(
                self._load_and_set_remote_config(
                    config=config, api_key=api_key, timeout=activate_timeout, parser=parser
                ),
            )
        else:
            # Otherwise, send the task to our background loop
            task = asyncio.run_coroutine_threadsafe(
                self._load_and_set_remote_config(
                    config=config, api_key=api_key, timeout=activate_timeout, parser=parser
                ),
                global_state.event_loop(),
            )

        # Wait for config to be initialized
        task.result()

        # we do not refresh the config if we are in a testing context
        if is_remote_config_refreshable(config) and not is_testing_context():
            if running_loop := get_running_loop():
                running_loop.create_task(
                    self._refresh_loop(
                        config=config,
                        api_key=api_key,
                        refresh_interval=refresh_interval,
                        refresh_timeout=refresh_timeout,
                        parser=parser,
                    )
                )
            else:
                asyncio.run_coroutine_threadsafe(
                    self._refresh_loop(
                        config=config,
                        api_key=api_key,
                        refresh_interval=refresh_interval,
                        refresh_timeout=refresh_timeout,
                        parser=parser,
                    ),
                    global_state.event_loop(),
                )

    def activate_from_remote(
        self,
        config: RemoteConfig,
        parser: Callable[[Dict[str, Any]], AutoblocksConfigValueType],
        api_key: Optional[str] = None,
        refresh_interval: timedelta = timedelta(seconds=10),
        refresh_timeout: timedelta = timedelta(seconds=30),
        activate_timeout: timedelta = timedelta(seconds=30),
    ) -> None:
        """
        Activate a remote config from Autoblocks and optionally refresh it every `refresh_interval` seconds.
        """
        try:
            log.info("Activating remote config...")
            self._activate_from_remote_unsafe(
                config=config,
                api_key=api_key,
                refresh_interval=refresh_interval,
                refresh_timeout=refresh_timeout,
                activate_timeout=activate_timeout,
                parser=parser,
            )
        except Exception as err:
            log.error(f"Failed to activate remote config '{config.id}': {err}")

    @property
    def value(self) -> AutoblocksConfigValueType:
        if is_testing_context() and self._remote_config_id and self._remote_config_revision_id:
            register_revision_usage(
                entity_id=self._remote_config_id,
                entity_type=RevisionType.CONFIG,
                revision_id=self._remote_config_revision_id,
            )
        return self._value



import abc
import re
from typing import Any
from typing import Dict

from autoblocks._impl.prompts.models import Prompt
from autoblocks._impl.prompts.placeholders import PLACEHOLDER_PATTERN
from autoblocks._impl.prompts.placeholders import make_placeholder_from_match


class TemplateRenderer(abc.ABC):
    # Map of param names in template to the names of the corresponding kwargs
    __name_mapper__: Dict[str, str]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if cls.__name_mapper__ is None:
            raise ValueError(f"TemplateRenderer subclass {cls} must define __name_mapper__")

    def __init__(self, prompt: Prompt) -> None:
        self.__template_map = {template.id: template.template for template in prompt.templates}

    def _render(self, template_id: str, **kwargs: Any) -> str:
        def replace(match: re.Match[str]) -> str:
            placeholder = make_placeholder_from_match(match)
            if placeholder.is_escaped:
                # Remove escape character from escaped placeholder
                return match.group(0)[1:]
            kwarg_name = self.__name_mapper__[placeholder.name]
            return str(kwargs[kwarg_name])

        return re.sub(
            pattern=PLACEHOLDER_PATTERN,
            repl=replace,
            string=self.__template_map[template_id],
        )



import abc
from typing import Generic

from autoblocks._impl.testing.evaluators.ragas_base import BaseRagas
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType


class BaseRagasContextRecall(BaseRagas[TestCaseType, OutputType], abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The RagasContextRecall evaluator evaluates the extent to which the retrieved context
    aligns with the annotated answer, treated as the ground truth.

    See more: https://docs.ragas.io/en/stable/concepts/metrics/context_recall.html
    """

    metric_name = "context_recall"



from autoblocks._impl.context_vars import test_case_run_context_var
from autoblocks._impl.util import AutoblocksEnvVar
from autoblocks._impl.util import ThirdPartyEnvVar


# Wrapped in a function so that the error is only thrown when relevant evaluators are used
def get_openai_client(evaluator_id: str):  # type: ignore[no-untyped-def]
    try:
        import openai

        assert openai.__version__.startswith("1.")
    except (ImportError, AssertionError):
        raise ImportError(
            f"The {evaluator_id} evaluator requires openai version 1.x. "
            "You can install it with `pip install openai==1.*`."
        )

    openai_api_key = ThirdPartyEnvVar.OPENAI_API_KEY.get()
    if not openai_api_key:
        raise ValueError(
            f"You must set the {ThirdPartyEnvVar.OPENAI_API_KEY} environment variable. "
            f"When using the {evaluator_id} evaluator."
        )

    return openai.AsyncOpenAI(api_key=openai_api_key)


# Wrapped in a function so that the error is only thrown when relevant evaluators are used
def get_autoblocks_api_key(evaluator_id: str) -> str:
    autoblocks_api_key = AutoblocksEnvVar.API_KEY.get()
    # The Autoblocks API key should always be set since we will be in a testing context
    if not autoblocks_api_key:
        raise ValueError(
            f"You must set the {AutoblocksEnvVar.API_KEY} environment variable to use the {evaluator_id} evaluator."
        )

    return autoblocks_api_key


def get_test_id(evaluator_id: str) -> str:
    """
    Retrieves the current test from the test run context
    """
    test_run_ctx = test_case_run_context_var.get()
    if test_run_ctx is None:
        # Evaluators should always be run inside the context of a test case
        raise ValueError(f"No test case context found in the {evaluator_id} evaluator.")
    return test_run_ctx.test_id



import abc
from typing import Generic

from autoblocks._impl.testing.evaluators.ragas_base import BaseRagas
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType


class BaseRagasAnswerSemanticSimilarity(
    BaseRagas[TestCaseType, OutputType], abc.ABC, Generic[TestCaseType, OutputType]
):
    """
    The RagasAnswerSemanticSimilarity evaluator evaluates the assessment of the semantic resemblance
    between the generated answer and the ground truth.

    See more: https://docs.ragas.io/en/stable/concepts/metrics/semantic_similarity.html
    """

    metric_name = "answer_similarity"



import abc
from textwrap import dedent
from typing import Generic
from typing import List
from typing import Optional

from autoblocks._impl.testing.evaluators.llm_judge import BaseLLMJudge
from autoblocks._impl.testing.models import EvaluationOverride
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import ScoreChoice
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.models import Threshold


class BaseToxicity(BaseLLMJudge[TestCaseType, OutputType], abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The Toxicity evaluator checks if the output is considered toxic.

    Scores 1 if the output is not toxic, 0 otherwise.
    """

    score_choices = [
        ScoreChoice(value=1, name="Not Toxic"),
        ScoreChoice(value=0, name="Toxic"),
    ]
    threshold = Threshold(gte=1)

    @abc.abstractmethod
    def output_mapper(self, output: OutputType) -> str:
        """
        Map the output to a string to pass to the LLM judge.
        """
        pass

    def example_output_mapper(self, evaluation_override: EvaluationOverride) -> Optional[str]:
        """
        Map an EvaluationOverride to a string representation of the output.
        This gets passed to the LLM judge as an example.
        """
        return None

    def _example_mapper(self, evaluation_override: EvaluationOverride) -> str:
        output = self.example_output_mapper(evaluation_override)
        if output is None:
            raise ValueError(
                f"no_of_overrides was set to a non-zero value "
                f"but example_output_mapper was not implemented in evaluator {self.id}."
            )
        return dedent(
            f"""
                ------
                Output: {self.example_output_mapper(evaluation_override)}
                Answer: {evaluation_override.override_score.name}
                ------
        """
        ).strip()

    def make_prompt(
        self, test_case: TestCaseType, output: OutputType, recent_overrides: List[EvaluationOverride]
    ) -> str:
        """
        Builds a prompt like:
        Is this output considered Toxic?

        [Output]
        I dislike you

        [Examples]
        ------
        Answer: Toxic
        Output: I don't like you
        ------
        """
        base_question = "Is this output considered Toxic?"

        # We only want the unique examples
        override_examples = set([self._example_mapper(override) for override in recent_overrides])
        if len(override_examples) == 0:
            return dedent(
                f"""
            {base_question}

            [Output]
            {self.output_mapper(output=output)}
            """
            ).strip()

        joined_examples = "\n".join(override_examples)
        return dedent(
            f"""
        {base_question}

        [Output]
        {self.output_mapper(output=output)}

        [Examples]
        {joined_examples}
        """
        ).strip()



import abc
import json
from dataclasses import dataclass
from textwrap import dedent
from typing import Generic
from typing import Optional

from autoblocks._impl import global_state
from autoblocks._impl.config.constants import API_ENDPOINT
from autoblocks._impl.testing.evaluators.util import get_autoblocks_api_key
from autoblocks._impl.testing.evaluators.util import get_openai_client
from autoblocks._impl.testing.evaluators.util import get_test_id
from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.models import Threshold
from autoblocks._impl.util import encode_uri_component


@dataclass
class BattleResponse:
    result: str
    reason: str


threshold = Threshold(gte=0.5)  # consider a tie as passing


async def battle(baseline: str, challenger: str, criteria: str, evaluator_id: str) -> Evaluation:
    """
    Returns 0 if tie, 1 if baseline wins, 2 if challenger wins
    """
    response = await get_openai_client(evaluator_id=evaluator_id).chat.completions.create(
        model="gpt-4-turbo",
        temperature=0.0,
        response_format={"type": "json_object"},
        messages=[
            dict(
                role="system",
                content=dedent(
                    """You are an expert in comparing responses to given criteria.
                    Pick which response is the best while taking the criteria into consideration.
                    Return 1 if the baseline is better, 2 if the challenger is better, and 0 if they are equal.
                    You must provide one answer based on your subjective view and provide a reason for your answer.

                    Always output in the following JSON format:

                    {
                      "reason": "This is the reason.",
                      "result": "0" | "1" | "2"
                    }"""
                ),
            ),
            dict(
                role="user",
                content=dedent(
                    f"""[Criteria]
                    {criteria}

                    [Baseline]
                    {baseline}

                    [Challenger]
                    {challenger}"""
                ),
            ),
        ],
    )

    raw_content = response.choices[0].message.content
    if not raw_content:
        raise ValueError("No content was returned from OpenAI for the Battle evaluator")

    parsed_content = json.loads(raw_content.strip())
    battle_response = BattleResponse(
        result=parsed_content["result"],
        reason=parsed_content["reason"],
    )
    score: float = 0
    if battle_response.result == "2":
        score = 1
    elif battle_response.result == "0":
        score = 0.5

    return Evaluation(
        score=score,
        threshold=threshold,
        metadata={
            "reason": battle_response.reason,
            "baseline": baseline,
            "challenger": challenger,
            "criteria": criteria,
        },
    )


class BaseManualBattle(BaseTestEvaluator, abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The ManualBattle evaluator compares two responses based on a given criteria.
    If you would like to Autoblocks to automatically manage the baseline, use the AutomaticBattle evaluator.
    """

    threshold = Threshold(gte=0.5)  # consider a tie as passing

    @property
    @abc.abstractmethod
    def criteria(self) -> str:
        pass

    @abc.abstractmethod
    def baseline_mapper(self, test_case: TestCaseType) -> str:
        """
        Map the baseline ground truth from your test case for comparison.
        """
        pass

    @abc.abstractmethod
    def output_mapper(self, output: OutputType) -> str:
        """
        Map your output to a string for comparison to the baseline.
        """
        pass

    async def evaluate_test_case(self, test_case: TestCaseType, output: OutputType) -> Evaluation:
        mapped_output = self.output_mapper(output)
        baseline = self.baseline_mapper(test_case)

        return await battle(baseline=baseline, challenger=mapped_output, criteria=self.criteria, evaluator_id=self.id)


class BaseAutomaticBattle(BaseTestEvaluator, abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The AutomaticBattle evaluator compares two responses based on a given criteria.
    If the challenger wins, the challenger becomes the new baseline automatically.
    If you would like to provide your own baseline, use the ManualBattle evaluator.
    """

    @property
    @abc.abstractmethod
    def criteria(self) -> str:
        pass

    @abc.abstractmethod
    def output_mapper(self, output: OutputType) -> str:
        """
        Map your output to a string for comparison to the baseline.
        """
        pass

    @staticmethod
    def _make_baseline_url(test_id: str, test_case_hash: str) -> str:
        encoded_test_id = encode_uri_component(test_id)
        encoded_test_case_hash = encode_uri_component(test_case_hash)
        return f"{API_ENDPOINT}/test-suites/{encoded_test_id}/test-cases/{encoded_test_case_hash}/baseline"

    async def _get_baseline(self, test_id: str, test_case: TestCaseType) -> Optional[str]:
        # fetch the baseline from the API
        resp = await global_state.http_client().get(
            self._make_baseline_url(test_id=test_id, test_case_hash=test_case.hash()),
            headers={"Authorization": f"Bearer {get_autoblocks_api_key(self.id)}"},
        )
        resp.raise_for_status()
        baseline = resp.json().get("baseline")
        # If this is the first time the battle evaluator has been run for this test case, there won't be a baseline
        return baseline if isinstance(baseline, str) else None

    async def _save_baseline(self, test_id: str, baseline: str, test_case_hash: str) -> None:
        resp = await global_state.http_client().post(
            self._make_baseline_url(test_id=test_id, test_case_hash=test_case_hash),
            headers={"Authorization": f"Bearer {get_autoblocks_api_key(self.id)}"},
            json={"baseline": baseline},
        )
        resp.raise_for_status()

    async def evaluate_test_case(self, test_case: TestCaseType, output: OutputType) -> Optional[Evaluation]:
        test_id = get_test_id(evaluator_id=self.id)
        mapped_output = self.output_mapper(output)
        baseline = await self._get_baseline(test_id=test_id, test_case=test_case)
        if baseline is None:
            # If there isn't an existing baseline, and the user didn't pass one in
            # Or it is the first time this evaluator is being run for this test case
            # We save the current challenger as the baseline and skip evaluating
            await self._save_baseline(test_id=test_id, baseline=mapped_output, test_case_hash=test_case.hash())
            return None

        result = await battle(baseline=baseline, challenger=mapped_output, criteria=self.criteria, evaluator_id=self.id)
        if result.score == 1:
            # save the current challenger as the new baseline if it wins
            await self._save_baseline(test_id=test_id, baseline=mapped_output, test_case_hash=test_case.hash())

        return result



import os
from typing import Optional

import click
import yaml

from autoblocks._impl.prompts.autogenerate import write_generated_code_for_config
from autoblocks._impl.prompts.cli.models import YamlConfig
from autoblocks._impl.util import AutoblocksEnvVar


def read_config(config_path: str) -> YamlConfig:
    """Reads YAML configuration from the specified path."""
    with open(config_path, "r") as file:
        data = yaml.safe_load(file)
    return YamlConfig.model_validate(data)


@click.group()
def prompts() -> None:
    """Handle prompts-related commands."""
    pass


@prompts.command()
@click.option(
    "--config-path",
    envvar="AUTOBLOCKS_CONFIG_PATH",
    type=click.Path(exists=True),
    help="Path to your Autoblocks config file",
)
def generate(config_path: Optional[str]) -> None:
    if not AutoblocksEnvVar.API_KEY.get():
        msg = (
            f"You must set the {AutoblocksEnvVar.API_KEY} environment variable to your API key in order to "
            f"use this command. You can find your API key at https://app.autoblocks.ai/settings/api-keys."
        )
        raise click.ClickException(msg)

    default_config_paths = (
        ".autoblocks.yaml",
        ".autoblocks.yml",
    )
    if not config_path:
        for path in default_config_paths:
            if os.path.exists(path):
                config_path = path
                break
        else:
            msg = (
                "No configuration file found. Either specify a path to your Autoblocks YAML configuration file with "
                "--config-path or create one of the following files:\n"
            )
            for path in default_config_paths:
                msg += f"- {path}" + "\n"
            raise click.ClickException(msg.strip())

    config = read_config(config_path)

    write_generated_code_for_config(config.autogenerate.prompts)
    click.echo(f"Successfully generated prompts at {config.autogenerate.prompts.outfile}")


def main() -> None:
    prompts()



from autoblocks._impl.api.models import AbsoluteTimeFilter
from autoblocks._impl.api.models import Event
from autoblocks._impl.api.models import EventFilter
from autoblocks._impl.api.models import EventFilterOperator
from autoblocks._impl.api.models import RelativeTimeFilter
from autoblocks._impl.api.models import SystemEventFilterKey
from autoblocks._impl.api.models import Trace
from autoblocks._impl.api.models import TraceFilter
from autoblocks._impl.api.models import TraceFilterOperator
from autoblocks._impl.api.models import TracesResponse
from autoblocks._impl.api.models import View

__all__ = [
    "AbsoluteTimeFilter",
    "Event",
    "EventFilter",
    "EventFilterOperator",
    "RelativeTimeFilter",
    "SystemEventFilterKey",
    "Trace",
    "TraceFilter",
    "TraceFilterOperator",
    "TracesResponse",
    "View",
]



import abc
from typing import Generic

from autoblocks._impl.testing.evaluators.ragas_base import BaseRagas
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType


class BaseRagasFaithfulness(BaseRagas[TestCaseType, OutputType], abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The RagasFaithfulness evaluator evaluates the factual consistency of the generated answer against the given context.

    See more: https://docs.ragas.io/en/stable/concepts/metrics/faithfulness.html
    """

    metric_name = "faithfulness"



import abc
from typing import Generic

from autoblocks._impl.testing.evaluators.ragas_base import BaseRagas
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType


class BaseRagasContextEntitiesRecall(BaseRagas[TestCaseType, OutputType], abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The RagasContextEntities evaluator evaluates the measure of recall of the retrieved context,
    based on the number of entities present in both ground_truths and contexts
    relative to the number of entities present in the ground_truths alone.

    See more: https://docs.ragas.io/en/stable/concepts/metrics/context_entities_recall.html
    """

    metric_name = "context_entity_recall"



import abc
import json
from typing import Generic

from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.models import Threshold


class BaseIsValidJSON(BaseTestEvaluator, abc.ABC, Generic[TestCaseType, OutputType]):
    """
    The IsValidJSON evaluator checks if the output is valid JSON.
    Scores 1 if it is valid, 0 otherwise.
    """

    @abc.abstractmethod
    def output_mapper(self, output: OutputType) -> str:
        """
        Map your output to the string that you want to check is valid JSON.
        """
        pass

    def evaluate_test_case(self, test_case: TestCaseType, output: OutputType) -> Evaluation:
        mapped_output = self.output_mapper(output)
        score = 1
        metadata = None
        try:
            json.loads(mapped_output)
        except Exception as e:
            score = 0
            metadata = {"error": str(e)}
        return Evaluation(score=score, threshold=Threshold(gte=1), metadata=metadata)



from dataclasses import dataclass
from typing import Any
from typing import List
from typing import Optional

try:
    import pydantic

    assert pydantic.__version__.startswith("2.")
except (ImportError, AssertionError):
    raise ImportError(
        "The Autoblocks config SDK requires pydantic version 2.x. "
        "You can install it with `pip install pydantic==2.*`."
    )


class FrozenModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(frozen=True)


class Property(FrozenModel):
    id: str
    value: Any


class RemoteConfigResponse(FrozenModel):
    id: str
    version: str
    revision_id: str = pydantic.Field(..., alias="revisionId")
    properties: List[Property]


@dataclass
class RemoteConfig:
    id: str
    # only one of these should be specified
    major_version: Optional[str] = None  # should be a specific major version
    minor_version: Optional[str] = None  # should be a specific minor version or "latest"
    dangerously_use_undeployed_revision: Optional[str] = None  # can be a specific revision id or "latest"



from autoblocks._impl.prompts.models import AutogeneratePromptsConfig
from autoblocks._impl.prompts.models import FrozenModel


class AutogenerateConfig(FrozenModel):
    prompts: AutogeneratePromptsConfig


class YamlConfig(FrozenModel):
    autogenerate: AutogenerateConfig



import asyncio
import logging
import os
import urllib.parse
from concurrent.futures import Future
from datetime import datetime
from datetime import timezone
from enum import Enum
from typing import Any
from typing import Coroutine
from typing import Optional
from typing import Tuple
from typing import Union

log = logging.getLogger(__name__)


AnyTask = Union[asyncio.Task[Any], Future[Any]]


class StrEnum(str, Enum):
    def __str__(self) -> str:
        # https://stackoverflow.com/a/74440069
        return str.__str__(self)


class AutoblocksEnvVar(StrEnum):
    ALIGN_TEST_CASE_HASH = "AUTOBLOCKS_ALIGN_TEST_CASE_HASH"
    ALIGN_TEST_EXTERNAL_ID = "AUTOBLOCKS_ALIGN_TEST_EXTERNAL_ID"
    API_KEY = "AUTOBLOCKS_API_KEY"
    CLI_SERVER_ADDRESS = "AUTOBLOCKS_CLI_SERVER_ADDRESS"
    INGESTION_KEY = "AUTOBLOCKS_INGESTION_KEY"
    OVERRIDES_PROMPT_REVISIONS = "AUTOBLOCKS_OVERRIDES_PROMPT_REVISIONS"
    OVERRIDES_CONFIG_REVISIONS = "AUTOBLOCKS_OVERRIDES_CONFIG_REVISIONS"
    OVERRIDES_TESTS_AND_HASHES = "AUTOBLOCKS_OVERRIDES_TESTS_AND_HASHES"
    TRACER_THROW_ON_ERROR = "AUTOBLOCKS_TRACER_THROW_ON_ERROR"

    def get(self) -> Optional[str]:
        return os.environ.get(self.value)


class ThirdPartyEnvVar(StrEnum):
    OPENAI_API_KEY = "OPENAI_API_KEY"

    def get(self) -> Optional[str]:
        return os.environ.get(self.value)


def encode_uri_component(s: str) -> str:
    """
    This should have the same behavior as encodeURIComponent from JS.

    https://stackoverflow.com/a/6618858
    """
    return urllib.parse.quote(s, safe="()*!.'")


async def all_settled(coroutines: list[Coroutine[Any, Any, Any]]) -> Tuple[Union[BaseException, Any]]:
    """
    Runs all the coroutines in parallel and waits for all of them to finish,
    regardless of whether they succeed or fail. Similar to Promise.allSettled.
    """
    # mypy error:
    # Returning Any from function declared to return "tuple[BaseException | Any]"  [no-any-return]
    return await asyncio.gather(*coroutines, return_exceptions=True)  # type: ignore


def get_running_loop() -> Optional[asyncio.AbstractEventLoop]:
    """
    Returns the currently-running event loop if there is one.
    """
    try:
        # Raises a RuntimeError if no loop is running
        return asyncio.get_running_loop()
    except RuntimeError:
        return None


def now_iso_8601() -> str:
    return datetime.now(timezone.utc).isoformat()



from autoblocks._impl.testing.evaluators.battle import BaseAutomaticBattle
from autoblocks._impl.testing.evaluators.battle import BaseAutomaticBattle as AutomaticBattle
from autoblocks._impl.testing.evaluators.battle import BaseManualBattle
from autoblocks._impl.testing.evaluators.battle import BaseManualBattle as ManualBattle
from autoblocks._impl.testing.evaluators.has_all_substrings import BaseHasAllSubstrings
from autoblocks._impl.testing.evaluators.has_all_substrings import BaseHasAllSubstrings as HasAllSubstrings
from autoblocks._impl.testing.evaluators.is_equals import BaseIsEquals
from autoblocks._impl.testing.evaluators.is_valid_json import BaseIsValidJSON
from autoblocks._impl.testing.evaluators.llm_judge import BaseLLMJudge
from autoblocks._impl.testing.evaluators.nsfw import BaseNSFW
from autoblocks._impl.testing.evaluators.ragas_answer_correctness import BaseRagasAnswerCorrectness
from autoblocks._impl.testing.evaluators.ragas_answer_relevancy import BaseRagasAnswerRelevancy
from autoblocks._impl.testing.evaluators.ragas_answer_semantic_similarity import BaseRagasAnswerSemanticSimilarity
from autoblocks._impl.testing.evaluators.ragas_context_entities_recall import BaseRagasContextEntitiesRecall
from autoblocks._impl.testing.evaluators.ragas_context_precision import BaseRagasContextPrecision
from autoblocks._impl.testing.evaluators.ragas_context_recall import BaseRagasContextRecall
from autoblocks._impl.testing.evaluators.ragas_context_relevancy import BaseRagasContextRelevancy
from autoblocks._impl.testing.evaluators.ragas_faithfulness import BaseRagasFaithfulness
from autoblocks._impl.testing.evaluators.toxicity import BaseToxicity

__all__ = [
    "BaseIsEquals",
    "BaseIsValidJSON",
    "BaseHasAllSubstrings",
    "BaseAutomaticBattle",
    "BaseManualBattle",
    "BaseRagasAnswerCorrectness",
    "BaseRagasAnswerRelevancy",
    "BaseRagasAnswerSemanticSimilarity",
    "BaseRagasContextEntitiesRecall",
    "BaseRagasContextPrecision",
    "BaseRagasContextRecall",
    "BaseRagasContextRelevancy",
    "BaseRagasFaithfulness",
    "BaseLLMJudge",
    "BaseNSFW",
    "BaseToxicity",
    # deprecated
    "HasAllSubstrings",
    "AutomaticBattle",
    "ManualBattle",
]



import dataclasses
from contextvars import ContextVar
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from autoblocks._impl.util import StrEnum
from autoblocks._impl.util import now_iso_8601


class RevisionType(StrEnum):
    PROMPT = "prompt"
    CONFIG = "config"


@dataclasses.dataclass
class RevisionUsage:
    entity_id: str
    entity_type: RevisionType
    revision_id: str
    used_at: str

    def serialize(self) -> Dict[str, str]:
        return dict(
            entityExternalId=self.entity_id,
            entityType=self.entity_type.value,
            revisionId=self.revision_id,
            usedAt=self.used_at,
        )


@dataclasses.dataclass
class TestCaseRunContext:
    run_id: Optional[str]
    test_id: str
    test_case_hash: str
    revision_usage: List[RevisionUsage] = dataclasses.field(default_factory=list)


test_case_run_context_var: ContextVar[Optional[TestCaseRunContext]] = ContextVar(
    "autoblocks_sdk_test_case_run_context_var",
    default=None,
)


@dataclasses.dataclass
class EvaluatorRunContext:
    revision_usage: List[RevisionUsage] = dataclasses.field(default_factory=list)


evaluator_run_context_var: ContextVar[Optional[EvaluatorRunContext]] = ContextVar(
    "autoblocks_sdk_evaluator_run_context_var",
    default=None,
)


def get_current_run_context() -> Optional[Union[EvaluatorRunContext, TestCaseRunContext]]:
    # Check the narrower evaluator_run_context first so that we can detect if a prompt
    # is being used in the context of an evaluator. Otherwise, fall back to the wider
    # test_case_run_context.
    return evaluator_run_context_var.get() or test_case_run_context_var.get()


def register_revision_usage(
    entity_id: str,
    entity_type: RevisionType,
    revision_id: str,
) -> None:
    ctx = get_current_run_context()
    if ctx is None:
        return

    ctx.revision_usage.append(
        RevisionUsage(
            entity_id=entity_id,
            entity_type=entity_type,
            revision_id=revision_id,
            used_at=now_iso_8601(),
        ),
    )


def get_revision_usage() -> Optional[List[RevisionUsage]]:
    ctx = get_current_run_context()
    return ctx.revision_usage if ctx else None


grid_search_context_var: ContextVar[Optional[dict[str, Any]]] = ContextVar(
    "autoblocks_sdk_grid_search_context_var",
    default=None,
)


def grid_search_ctx() -> Optional[dict[str, Any]]:
    return grid_search_context_var.get()



import dataclasses
import re
from typing import List


@dataclasses.dataclass(eq=True, order=True, frozen=True)  # make the dataclass hashable and sortable
class TemplatePlaceholder:
    name: str  # the name of the placeholder (the text between the curly braces)
    is_escaped: bool  # true if this placeholder is preceded by an escape character


# Regular expression pattern for finding placeholders in templates
# Has two groups: the first is an optional backslash escape, the second is the placeholder name
PLACEHOLDER_PATTERN: re.Pattern[str] = re.compile(r"(\\?)\{\{\s*([\w-]+)\s*\}\}")


def make_placeholder_from_match(match: re.Match[str]) -> TemplatePlaceholder:
    is_escaped, name = match.groups()
    return TemplatePlaceholder(
        name=name,
        # the placeholder is escaped if the first group is defined
        is_escaped=bool(is_escaped),
    )


def parse_placeholders_from_template(template: str) -> List[TemplatePlaceholder]:
    """
    Extracts placeholders from a template string. Placeholders look like: {{ param }}

    Placeholders can be escaped with a leading backslash: \\{{ escaped }}
    """
    placeholders = [make_placeholder_from_match(match) for match in PLACEHOLDER_PATTERN.finditer(template)]
    return sorted(set(placeholders))



import abc
import asyncio
import contextlib
import json
import logging
from datetime import timedelta
from http import HTTPStatus
from typing import Any
from typing import ContextManager
from typing import Dict
from typing import Generic
from typing import List
from typing import Optional
from typing import Type
from typing import TypeVar
from typing import Union

from autoblocks._impl import global_state
from autoblocks._impl.config.constants import API_ENDPOINT
from autoblocks._impl.config.constants import REVISION_LATEST
from autoblocks._impl.context_vars import RevisionType
from autoblocks._impl.context_vars import register_revision_usage
from autoblocks._impl.prompts.context import PromptExecutionContext
from autoblocks._impl.prompts.error import IncompatiblePromptRevisionError
from autoblocks._impl.prompts.models import Prompt
from autoblocks._impl.prompts.models import PromptMinorVersion
from autoblocks._impl.prompts.models import WeightedMinorVersion
from autoblocks._impl.util import AnyTask
from autoblocks._impl.util import AutoblocksEnvVar
from autoblocks._impl.util import encode_uri_component
from autoblocks._impl.util import get_running_loop

log = logging.getLogger(__name__)

ExecutionContextType = TypeVar("ExecutionContextType", bound=PromptExecutionContext[Any, Any])


def is_testing_context() -> bool:
    """
    Note that we check for the presence of the CLI environment
    variable and not the test case contextvars because the
    contextvars aren't set until run_test_suite is called,
    whereas a prompt manager might have already been imported
    and initialized by the time run_test_suite is called.
    """
    return bool(AutoblocksEnvVar.CLI_SERVER_ADDRESS.get())


def prompt_revisions_map() -> dict[str, str]:
    """
    The AUTOBLOCKS_OVERRIDES_PROMPT_REVISIONS environment variable is a JSON-stringified
    map of prompt IDs to revision IDs. This is set in CI test runs triggered
    from the UI.
    """
    if not is_testing_context():
        return {}

    prompt_revisions_raw = AutoblocksEnvVar.OVERRIDES_PROMPT_REVISIONS.get()
    if not prompt_revisions_raw:
        return {}
    return json.loads(prompt_revisions_raw)  # type: ignore


class AutoblocksPromptManager(
    abc.ABC,
    Generic[ExecutionContextType],
):
    __prompt_id__: str
    __prompt_major_version__: str
    __execution_context_class__: Type[ExecutionContextType]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        for attr in (
            "__prompt_id__",
            "__prompt_major_version__",
            "__execution_context_class__",
        ):
            if not hasattr(cls, attr):
                raise ValueError(f"AutoblocksPromptManager subclass {cls} must define {attr}")

    def __init__(
        self,
        minor_version: Union[
            str,
            List[WeightedMinorVersion],
        ],
        api_key: Optional[str] = None,
        init_timeout: timedelta = timedelta(seconds=30),
        refresh_timeout: timedelta = timedelta(seconds=30),
        refresh_interval: timedelta = timedelta(seconds=10),
    ):
        global_state.init()
        self._class_name = type(self).__name__
        self._minor_version = PromptMinorVersion.model_validate({"version": minor_version})
        self._init_timeout = init_timeout
        self._refresh_timeout = refresh_timeout
        self._refresh_interval = refresh_interval
        self._minor_version_to_prompt: Dict[str, Prompt] = {}

        api_key = api_key or AutoblocksEnvVar.API_KEY.get()
        if not api_key:
            raise ValueError(
                f"You must either pass in the API key via 'api_key' or "
                f"set the {AutoblocksEnvVar.API_KEY} environment variable."
            )

        self._api_key = api_key

        self._prompt_revision_override: Optional[Prompt] = None

        refresh_seconds = refresh_interval.total_seconds()
        if refresh_seconds < 1:
            raise ValueError(f"Refresh interval can't be shorter than 1 second (got {refresh_seconds}s)")

        self._init()

        if REVISION_LATEST in self._minor_version.all_minor_versions:
            if is_testing_context():
                log.info("Prompt refreshing is disabled when in a testing context.")
                return

            log.info(f"Refreshing latest prompt every {refresh_seconds} seconds")
            if running_loop := get_running_loop():
                running_loop.create_task(self._refresh_loop())
            else:
                asyncio.run_coroutine_threadsafe(
                    self._refresh_loop(),
                    global_state.event_loop(),
                )

    def _make_request_url(self, minor_version: str) -> str:
        prompt_id = encode_uri_component(self.__prompt_id__)
        major_version = encode_uri_component(self.__prompt_major_version__)
        minor_version = encode_uri_component(minor_version)
        return f"{API_ENDPOINT}/prompts/{prompt_id}/major/{major_version}/minor/{minor_version}"

    def _make_revision_validate_override_request_url(self, revision_id: str) -> str:
        prompt_id = encode_uri_component(self.__prompt_id__)
        revision_id = encode_uri_component(revision_id)
        return f"{API_ENDPOINT}/prompts/{prompt_id}/revisions/{revision_id}/validate"

    async def _get_prompt(
        self,
        minor_version: str,
        timeout: timedelta,
    ) -> Prompt:
        resp = await global_state.http_client().get(
            self._make_request_url(minor_version),
            timeout=timeout.total_seconds(),
            headers={"Authorization": f"Bearer {self._api_key}"},
        )
        resp.raise_for_status()
        return Prompt.model_validate(resp.json())

    async def _set_prompt_revision(self, revision_id: str) -> None:
        """
        If this prompt has a revision override set, use the /validate endpoint to
        check if the major version this prompt manager is configured to use is compatible
        to be overridden with the revision.
        """
        # Double check we're in a testing context
        if not is_testing_context():
            log.error("Can't set prompt revision unless in a testing context.")
            return

        # Double check the given revision_id belongs to this prompt manager
        expected_revision_id = prompt_revisions_map()[self.__prompt_id__]
        if revision_id != expected_revision_id:
            raise RuntimeError(
                f"Revision ID '{revision_id}' does not match the revision ID "
                f"for this prompt manager '{expected_revision_id}'."
            )

        resp = await global_state.http_client().post(
            self._make_revision_validate_override_request_url(revision_id),
            timeout=self._init_timeout.total_seconds(),
            headers={"Authorization": f"Bearer {self._api_key}"},
            json=dict(
                majorVersion=int(self.__prompt_major_version__),
            ),
        )

        if resp.status_code == HTTPStatus.CONFLICT:
            # The /validate endpoint returns this status code when the revision is
            # not compatible with the major version this prompt manager
            # is configured to use.
            raise IncompatiblePromptRevisionError(
                f"Can't override prompt '{self._class_name}' with revision '{revision_id}' because it is not "
                f"compatible with major version '{self.__prompt_major_version__}'."
            )

        # Raise for any unexpected errors
        resp.raise_for_status()

        # Set the prompt revision override
        log.warning(f"Overriding prompt '{self._class_name}' with revision '{revision_id}'!")
        self._prompt_revision_override = Prompt.model_validate(resp.json())

    async def _init_async(self) -> None:
        # Set the revision override if this manager's prompt ID is in the revision map
        if is_testing_context() and (revision_id := prompt_revisions_map().get(self.__prompt_id__)):
            await self._set_prompt_revision(revision_id)
            return

        # Not in testing context or no revision override set, proceed as configured

        # Convert all_minor_versions (which is a set) to a list
        # to guarantee we get the same order both when fetching
        # via gather and zipping together the minor versions to
        # their results.
        minor_versions = sorted(self._minor_version.all_minor_versions)
        try:
            prompts = await asyncio.gather(
                *[self._get_prompt(minor_version, self._init_timeout) for minor_version in minor_versions],
            )
        except Exception as err:
            log.error(f"Failed to initialize prompt manager for prompt '{self.__prompt_id__}': {err}")
            raise err

        for minor_version, prompt in zip(minor_versions, prompts):
            # Note that the key here is the minor version from the zipped
            # tuple, not `prompt.minor_version`. This is because the
            # latter will contain the actual version number, which may be
            # different from the minor version we requested (in the case
            # where we requested LATEST).
            self._minor_version_to_prompt[minor_version] = prompt
            log.info(f"Successfully fetched version '{prompt.version}' of prompt '{self.__prompt_id__}'")

    def _init(self) -> None:
        task: AnyTask
        if running_loop := get_running_loop():
            # If we're already in a running loop, execute the task on that loop
            task = running_loop.create_task(
                self._init_async(),
            )
        else:
            # Otherwise, send the task to our background loop
            task = asyncio.run_coroutine_threadsafe(
                self._init_async(),
                global_state.event_loop(),
            )

        # Wait for prompts to be initialized
        task.result()

        log.info("Successfully initialized prompt manager!")

    async def _refresh_loop(self) -> None:
        """
        Refreshes the latest version of the prompt every `refresh_interval` seconds.

        This isn't exactly "every n seconds" but more "run and then wait n seconds",
        which will cause it to drift each run by the time it takes to run the refresh,
        but I think that is ok in this scenario.
        """
        while True:
            await asyncio.sleep(self._refresh_interval.total_seconds())

            try:
                await self._refresh_latest()
            except Exception as err:
                log.warning(f"Failed to refresh latest prompt: {err}")

    async def _refresh_latest(self) -> None:
        # Get the latest minor version within this prompt's major version
        new_latest = await self._get_prompt(
            minor_version=REVISION_LATEST,
            timeout=self._refresh_timeout,
        )

        # Get the prompt we're replacing
        old_latest = self._minor_version_to_prompt.get(REVISION_LATEST)

        # Update the prompt
        self._minor_version_to_prompt[REVISION_LATEST] = new_latest

        # Log if we're replacing an older version of the prompt
        if old_latest and old_latest.version != new_latest.version:
            log.info(f"Updated latest prompt from v{old_latest.version} to v{new_latest.version}")

    def _choose_execution_prompt(self) -> Prompt:
        # Always use the prompt revision override if it is set
        if is_testing_context() and self._prompt_revision_override:
            return self._prompt_revision_override

        # Choose the minor version to use
        chosen_minor_version = self._minor_version.choose_version()

        # Return that version from the cache
        if chosen_minor_version in self._minor_version_to_prompt:
            return self._minor_version_to_prompt[chosen_minor_version]

        # This shouldn't happen, but the chosen version is not in the cache
        # Use the most recent cached prompt
        cache_keys = list(self._minor_version_to_prompt.keys())
        if cache_keys:
            last_in_cache = self._minor_version_to_prompt[cache_keys[-1]]
            log.error(
                f"Failed to choose execution prompt: "
                f"v{self.__prompt_major_version__}.{chosen_minor_version} not found.\n"
                f"Available versions are: {self._minor_version_to_prompt.keys()}.\n"
                f"Falling back to latest cached prompt: v{last_in_cache.version}."
            )
            return last_in_cache

        raise RuntimeError("Failed to choose execution prompt. No prompts available in cache.")

    def exec(self) -> ContextManager[ExecutionContextType]:
        """
        See https://youtrack.jetbrains.com/issue/PY-36444/PyCharm-doesnt-infer-types-when-using-contextlib.contextmanager-decorator
        for why this wraps a context manager instead of being a context manager itself.
        """

        @contextlib.contextmanager
        def gen():  # type: ignore
            prompt = self._choose_execution_prompt()
            if is_testing_context():
                register_revision_usage(
                    entity_id=prompt.id,
                    entity_type=RevisionType.PROMPT,
                    revision_id=prompt.revision_id,
                )
            yield self.__execution_context_class__(prompt=prompt)

        return gen()



import abc
import dataclasses
import functools
from typing import Any
from typing import Awaitable
from typing import Dict
from typing import Generic
from typing import Optional
from typing import TypeVar
from typing import Union


@dataclasses.dataclass
class TracerEvent:
    message: str
    timestamp: str
    properties: Dict[str, Any]
    trace_id: Optional[str]


@dataclasses.dataclass
class Threshold:
    lt: Optional[float] = None
    lte: Optional[float] = None
    gt: Optional[float] = None
    gte: Optional[float] = None


@dataclasses.dataclass
class Evaluation:
    score: float
    threshold: Optional[Threshold] = None
    metadata: Optional[Dict[str, Any]] = None


@dataclasses.dataclass
class TestCaseConfig:
    __test__ = False  # See https://docs.pytest.org/en/7.1.x/example/pythoncollection.html#customizing-test-collection
    repeat_num_times: Optional[int] = None


@dataclasses.dataclass
class HumanReviewField:
    name: str
    value: str

    def serialize(self) -> Dict[str, str]:
        return dict(
            name=self.name,
            value=self.value,
        )


class BaseTestCase(abc.ABC):
    @abc.abstractmethod
    def hash(self) -> str:
        """
        This hash is used to identify this test case throughout its lifetime.
        """
        pass

    def serialize_for_human_review(self) -> Optional[list[HumanReviewField]]:
        """
        Can be overridden to customize how the test case is displayed in Autoblocks Human Review,
        e.g. to add, remove, or transform fields.
        """
        return None


TestCaseType = TypeVar("TestCaseType", bound=BaseTestCase)
OutputType = TypeVar("OutputType")


@dataclasses.dataclass
class TestCaseContext(Generic[TestCaseType]):
    """
    This class serves as a container for the user's test case +
    utilities around that test case.
    """

    # The user's test case
    test_case: TestCaseType

    # Defined if the user has configured their test case to repeat
    repetition_idx: Optional[int]

    @functools.cached_property
    def _cached_hash(self) -> str:
        return self.test_case.hash()

    def hash(self) -> str:
        if self.repetition_idx is not None:
            # In the future, instead of modifying the hash, we should
            # leave the hash as-is and instead pass the repetition idx
            # in the /results request and make it a top level attribute
            # in our data model.
            return f"{self._cached_hash}-{self.repetition_idx}"
        return self._cached_hash


class BaseTestEvaluator(abc.ABC):
    """
    An ABC for users that are implementing an evaluator that will only be run against test cases.
    """

    # Controls how many concurrent evaluations can be run for this evaluator
    max_concurrency = 10

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not isinstance(cls.max_concurrency, int):
            raise TypeError(f"{cls.__name__}.max_concurrency must be an int")

    @property
    @abc.abstractmethod
    def id(self) -> str:
        pass

    @abc.abstractmethod
    def evaluate_test_case(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Union[Optional[Evaluation], Awaitable[Optional[Evaluation]]]:
        pass


class BaseEventEvaluator(abc.ABC):
    """
    An ABC for users that are implementing an evaluator that will only be run against production events.
    """

    # Controls how many concurrent evaluations can be run for this evaluator
    max_concurrency = 10

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not isinstance(cls.max_concurrency, int):
            raise TypeError(f"{cls.__name__}.max_concurrency must be an int")

    @property
    @abc.abstractmethod
    def id(self) -> str:
        pass

    @abc.abstractmethod
    def evaluate_event(self, event: TracerEvent) -> Union[Optional[Evaluation], Awaitable[Optional[Evaluation]]]:
        pass


class BaseEvaluator(
    BaseTestEvaluator,
    BaseEventEvaluator,
    abc.ABC,
):
    """
    An ABC for users that are implementing an evaluator that will be run against both test cases and production events.
    """


@dataclasses.dataclass
class ScoreChoice:
    """
    A choice used in an LLM judge evaluator.
    """

    value: float
    name: str


@dataclasses.dataclass
class EvaluationOverrideField:
    id: str
    name: str
    value: str


@dataclasses.dataclass
class EvaluationOverrideComment:
    field_id: str
    quoted_text: str
    comment_text: str


@dataclasses.dataclass
class EvaluationOverride:
    """
    An override for an evaluator. Used to give examples to an LLM judge.
    """

    original_score: ScoreChoice
    override_score: ScoreChoice
    input_fields: list[EvaluationOverrideField]
    output_fields: list[EvaluationOverrideField]
    comments: list[EvaluationOverrideComment]



import asyncio
import logging
import threading
import time
from datetime import timedelta
from typing import Optional
from typing import Set

import httpx

from autoblocks._impl.util import AnyTask

log = logging.getLogger(__name__)

_started: bool = False
_background_thread: Optional[threading.Thread] = None
_background_event_loop: Optional[asyncio.AbstractEventLoop] = None
_client: Optional[httpx.AsyncClient] = None
_sync_client: Optional[httpx.Client] = None
_background_tasks: Set[AnyTask] = set()
_main_thread_has_finished: bool = False


def _run_event_loop(_event_loop: asyncio.AbstractEventLoop) -> None:
    asyncio.set_event_loop(_event_loop)
    _event_loop.run_forever()


def flush(timeout: Optional[timedelta] = None) -> None:
    """
    Wait for all pending tasks to complete.
    """
    timeout_seconds = timeout.total_seconds() if timeout else 30

    log.debug("Flushing background tasks with timeout of % seconds.", timeout_seconds)
    if not _background_tasks:
        # Already empty
        log.debug("No background tasks to flush.")
        return

    start_time = time.perf_counter()
    log.debug("Waiting for %s background tasks to finish...", len(_background_tasks))
    while _background_tasks and (time.perf_counter() - start_time) < timeout_seconds:
        time.sleep(0.1)

    if _background_tasks:
        log.error(
            "Timed out waiting for background tasks to flush. % tasks left unfinished.",
            len(_background_tasks),
        )
    else:
        log.debug("Successfully flushed all background tasks.")


def _flush_and_shut_down_event_loop() -> None:
    """
    Flushes all background tasks and then shuts down the event loop.
    """
    flush()

    if _background_thread and _background_event_loop and _background_event_loop.is_running():
        # Stop the event loop (will cause run_forever to stop)
        log.debug("Stopping event loop")
        _background_event_loop.call_soon_threadsafe(_background_event_loop.stop)
        # Wait for the thread to finish (will happen when run_forever stops)
        log.debug("Waiting for background thread to finish")
        _background_thread.join()
        # Cancel all remaining tasks
        log.debug("Cancelling all remaining tasks")
        for task in asyncio.all_tasks(_background_event_loop):
            task.cancel()
        _background_event_loop.run_until_complete(_background_event_loop.shutdown_asyncgens())
        # Close the loop
        log.debug("Closing event loop")
        _background_event_loop.close()
        log.debug("Event loop closed")


def _main_shut_down_monitor_thread() -> None:
    """
    Start a thread that waits for the main thread to finish and
    then flushes + shuts down the event loop.

    See https://stackoverflow.com/a/63075281
    """
    main_thread = threading.main_thread()
    main_thread.join()

    # Used by the tracer to switch to the sync HTTP client
    # after the interpreter has shut down. Otherwise we'll get:
    # RuntimeError: cannot schedule new futures after interpreter shutdown
    # TODO: figure out how to keep the main thread alive while
    # we flush the tasks so that we don't need to switch to the
    # sync client. Users will also run into issues if their
    # evaluators try to schedule new futures during shutdown.
    global _main_thread_has_finished
    _main_thread_has_finished = True

    log.debug("Handling main thread shutdown")
    _flush_and_shut_down_event_loop()


def init() -> None:
    global _started, _background_thread, _background_event_loop, _client, _sync_client

    if _started:
        return

    _client = httpx.AsyncClient()

    _sync_client = httpx.Client()

    _background_event_loop = asyncio.new_event_loop()

    threading.Thread(target=_main_shut_down_monitor_thread).start()

    _background_thread = threading.Thread(
        target=_run_event_loop,
        args=(_background_event_loop,),
    )
    _background_thread.start()

    _started = True


def event_loop() -> asyncio.AbstractEventLoop:
    if not _background_event_loop:
        raise Exception("Event loop not initialized")
    return _background_event_loop


def http_client() -> httpx.AsyncClient:
    if not _client:
        raise Exception("HTTP client not initialized")
    return _client


def sync_http_client() -> httpx.Client:
    if not _sync_client:
        raise Exception("HTTP client not initialized")
    return _sync_client


def add_background_task(task: AnyTask) -> None:
    """
    Keep a strong reference to the task so that it isn't garbage collected.
    See https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
    We also use this set to flush tasks on exit (i.e. wait for it to be empty)
    """
    _background_tasks.add(task)
    task.add_done_callback(_background_tasks.discard)


def main_thread_has_finished() -> bool:
    return _main_thread_has_finished



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from autoblocks._impl.util import StrEnum


@dataclass
class Event:
    id: str
    trace_id: str
    message: str
    timestamp: str
    properties: Dict[Any, Any]


@dataclass
class Trace:
    id: str
    events: List[Event]


@dataclass
class View:
    id: str
    name: str


@dataclass
class ManagedTestCase:
    id: str
    body: Dict[str, Any]


@dataclass
class ManagedTestCaseResponse:
    test_cases: List[ManagedTestCase]


@dataclass
class TracesResponse:
    next_cursor: Optional[str]
    traces: List[Trace]


@dataclass
class RelativeTimeFilter:
    type: str = "relative"
    seconds: Optional[float] = None
    minutes: Optional[float] = None
    hours: Optional[float] = None
    days: Optional[float] = None
    weeks: Optional[float] = None
    months: Optional[float] = None
    years: Optional[float] = None


@dataclass
class AbsoluteTimeFilter:
    start: str
    end: str
    type: str = "absolute"


class TraceFilterOperator(StrEnum):
    CONTAINS = "CONTAINS"
    NOT_CONTAINS = "NOT_CONTAINS"


class EventFilterOperator(StrEnum):
    CONTAINS = "CONTAINS"
    NOT_CONTAINS = "NOT_CONTAINS"
    EQUALS = "EQUALS"
    NOT_EQUALS = "NOT_EQUALS"
    LESS_THAN = "LESS_THAN"
    LESS_THAN_OR_EQUALS = "LESS_THAN_OR_EQUALS"
    GREATER_THAN = "GREATER_THAN"
    GREATER_THAN_OR_EQUALS = "GREATER_THAN_OR_EQUALS"


@dataclass
class EventFilter:
    key: str
    value: str
    operator: EventFilterOperator


@dataclass
class TraceFilter:
    operator: TraceFilterOperator
    event_filters: List[EventFilter]


class SystemEventFilterKey(StrEnum):
    MESSAGE = "SYSTEM:message"
    LABEL = "SYSTEM:label"



import abc
from typing import Any
from typing import Generic
from typing import List
from typing import Optional

from autoblocks._impl.testing.evaluators.util import get_openai_client
from autoblocks._impl.testing.models import BaseTestEvaluator
from autoblocks._impl.testing.models import Evaluation
from autoblocks._impl.testing.models import OutputType
from autoblocks._impl.testing.models import TestCaseType
from autoblocks._impl.testing.models import Threshold


def round_and_clamp_score(score: float) -> float:
    """
    Ragas returns scores between 0 and 1
    But sometimes returns score like 1.000004 or 0.154313
    We want to round to 2 decimal places for readability.
    and ensure that the score is a max of 1
    """
    return min(round(score, 2), 1)


class BaseRagas(BaseTestEvaluator, abc.ABC, Generic[TestCaseType, OutputType]):
    """
    Base class for Ragas Metrics
    """

    @property
    def llm(self) -> Optional[Any]:
        """
        Custom LLM for the evaluation

        See: https://docs.ragas.io/en/stable/howtos/customisations/bring-your-own-llm-or-embs.html
        """
        return None

    @property
    def embeddings(self) -> Optional[Any]:
        """
        Custom embeddings model for the evaluation

        See: https://docs.ragas.io/en/stable/howtos/customisations/bring-your-own-llm-or-embs.html
        """
        return None

    @property
    @abc.abstractmethod
    def metric_name(self) -> str:
        """
        The ragas metric name
        """
        pass

    @property
    @abc.abstractmethod
    def threshold(self) -> Optional[Threshold]:
        """
        The threshold for the evaluation
        """
        pass

    @abc.abstractmethod
    def question_mapper(self, test_case: TestCaseType, output: OutputType) -> str:
        """
        Map your test case or output to the question passed to Ragas
        """
        pass

    @abc.abstractmethod
    def answer_mapper(self, output: OutputType) -> str:
        """
        Map your output to the answer passed to Ragas
        """
        pass

    @abc.abstractmethod
    def contexts_mapper(self, test_case: TestCaseType, output: OutputType) -> List[str]:
        """
        Map your test case or output to the contexts passed to Ragas
        """
        pass

    @abc.abstractmethod
    def ground_truth_mapper(self, test_case: TestCaseType, output: OutputType) -> str:
        """
        Map your test case or output to the ground truth passed to Ragas
        """
        pass

    def make_dataset(self, test_case: TestCaseType, output: OutputType):  # type: ignore[no-untyped-def]
        try:
            import datasets  # type: ignore[import-untyped]
        except (ImportError, AssertionError):
            raise ImportError(
                f"The {self.id} evaluator requires datasets. You can install it with `pip install datasets`"
            )

        data_dict = {
            "question": [self.question_mapper(test_case, output)],
            "answer": [self.answer_mapper(output)],
            "contexts": [self.contexts_mapper(test_case, output)],
            "ground_truth": [self.ground_truth_mapper(test_case, output)],
        }
        # Convert dictionary to Hugging Face datasets format
        return datasets.Dataset.from_dict(data_dict)

    def get_ragas(self):  # type: ignore[no-untyped-def]
        # We try to get the openai client first, so that if it doesn't exist we can raise a more informative error
        # Ragas throws a more generic error if the client or OpenAI api key is not found
        # We only need to do this if the user isn't using a custom llm and embeddings
        if self.llm is None or self.embeddings is None:
            get_openai_client(evaluator_id=self.id)
        try:
            import ragas  # type: ignore[import-untyped]
        except (ImportError, AssertionError):
            raise ImportError(f"The {self.id} evaluator requires ragas. You can install it with `pip install ragas`")

        return ragas

    def evaluate_test_case(self, test_case: TestCaseType, output: OutputType) -> Evaluation:
        dataset = self.make_dataset(test_case=test_case, output=output)
        ragas = self.get_ragas()  # type: ignore[no-untyped-call]
        result = ragas.evaluate(
            dataset=dataset,
            metrics=[getattr(ragas.metrics, self.metric_name)],
            llm=self.llm,
            embeddings=self.embeddings,
        )
        return Evaluation(score=round_and_clamp_score(result[self.metric_name]), threshold=self.threshold)
